"""A Test Blueprint for testing Cookie Params.

The corresponding test can be found in tests/test_inbound/test_cookie
"""

from flask_jeroboam import Blueprint, Cookie

router = Blueprint("cookies_params_router", __name__, tags=["Cookies"])


@router.get("/cookie/int")
def get_cookie_as_int(cookie: int = Cookie()):
    return {"cookie": cookie}


@router.get("/cookie/str")
def get_cookie_as_str(cookie: str = Cookie()):
    return {"cookie": cookie}



import pytest
from flask.testing import FlaskClient

from flask_jeroboam import Blueprint, Jeroboam
from flask_jeroboam.models import InboundModel
from flask_jeroboam.view_arguments.functions import Body

response_not_valid_int = {
    "detail": [
        {
            "loc": ["body", "payload"],
            "msg": "Input should be a valid integer, unable to parse string as an integer",
            "type": "int_parsing",
        }
    ]
}


def _valid(value) -> dict:
    """Valid function."""
    return {"payload": value}


@pytest.mark.parametrize(
    "url,body_value,expected_status,expected_response",
    [
        ("/body/str", {"payload": "foobar"}, 201, _valid("foobar")),
        ("/body/int", {"payload": 123}, 201, _valid(123)),
        ("/body/int", {"payload": "not_a_valid_int"}, 400, response_not_valid_int),
        (
            "/body/base_model",
            {"page": 1, "type": "item"},
            201,
            {"page": 1, "type": "item"},
        ),
    ],
)
def test_post_body_operations(
    client: FlaskClient,
    url: str,
    body_value: dict,
    expected_status: int,
    expected_response: dict,
):
    """Testing Various GET operations with query parameters.

    GIVEN a GET endpoint configiured with query parameters
    WHEN a request is made to the endpoint
    THEN the request is parsed and validated accordingly
    """
    response = client.post(url, json=body_value)
    assert response.json == expected_response
    assert response.status_code == expected_status


def test_post_body_list_of_base_model(
    one_shot_app: Jeroboam, one_shot_client: FlaskClient
):
    """Test Body Parameter with POST method."""

    class InBound(InboundModel):
        """Inbound model."""

        item: str
        count: int

    @one_shot_app.post("/body/list_non_scalar", response_model=list[InBound])
    def post_body_list_non_scalar(payload: list[InBound] = Body()):
        return payload

    response = one_shot_client.post(
        "/body/list_non_scalar",
        json=[{"item": "foobar", "count": 1}, {"item": "bar", "count": 3}],
    )
    assert response.json == [
        {"item": "foobar", "count": 1},
        {"item": "bar", "count": 3},
    ]
    assert response.status_code == 201


def test_put_body_with_empty_string_rule(
    one_shot_app: Jeroboam, one_shot_client: FlaskClient
):
    """Regression test for #110: body field fails to build when URL rule is empty string."""
    bp = Blueprint("wines", __name__, url_prefix="/wines")

    @bp.put("")
    def put_empty_rule(payload: int):
        return {"payload": payload}

    one_shot_app.register_blueprint(bp)
    response = one_shot_client.put("/wines", json={"payload": 42})
    assert response.status_code == 201
    assert response.json == {"payload": 42}


def test_put_multi_body_with_empty_string_rule(
    one_shot_app: Jeroboam, one_shot_client: FlaskClient
):
    """Regression test for #110: multi-arg body (create_model path) with empty string rule."""
    bp = Blueprint("spirits", __name__, url_prefix="/spirits")

    @bp.put("")
    def put_empty_rule_multi(qty: int, name: str):
        return {"qty": qty, "name": name}

    one_shot_app.register_blueprint(bp)
    response = one_shot_client.put("/spirits", json={"qty": 5, "name": "cognac"})
    assert response.status_code == 201
    assert response.json == {"qty": 5, "name": "cognac"}



import pytest

not_a_valid_int = {
    "detail": [
        {
            "loc": ["cookie", "cookie"],
            "msg": "Input should be a valid integer, unable to parse string as an integer",
            "type": "int_parsing",
        }
    ]
}


def _valid(x):
    return {"cookie": x}


@pytest.mark.parametrize(
    "url,cookie_value,expected_status,expected_response",
    [
        ("/cookie/str", "foobar", 200, _valid("foobar")),
        ("/cookie/int", "123", 200, _valid(123)),
        ("/cookie/int", "not_valid_int", 400, not_a_valid_int),
    ],
)
def test_get_cookie(client, url, cookie_value, expected_status, expected_response):
    """Test Cookie Parameter with GET method.


    TODO: Allow Configuration of the returned Status Code.
    """
    client.set_cookie("cookie", cookie_value)
    response = client.get(url)
    assert response.status_code == expected_status
    assert response.json == expected_response



"""Testing path operations."""

import pytest

response_not_valid_bool = {
    "detail": [
        {
            "loc": ["path", "item_id"],
            "msg": "Input should be a valid boolean, unable to interpret input",
            "type": "bool_parsing",
        }
    ]
}

response_not_valid_int = {
    "detail": [
        {
            "loc": ["path", "item_id"],
            "msg": "Input should be a valid integer, unable to parse string as an integer",
            "type": "int_parsing",
        }
    ]
}

response_not_valid_float = {
    "detail": [
        {
            "loc": ["path", "item_id"],
            "msg": "Input should be a valid number, unable to parse string as a number",
            "type": "float_parsing",
        }
    ]
}

response_at_least_3 = {
    "detail": [
        {
            "loc": ["path", "item_id"],
            "msg": "String should have at least 3 characters",
            "type": "string_too_short",
            "ctx": {"min_length": 3},
        }
    ]
}


response_at_least_2 = {
    "detail": [
        {
            "loc": ["path", "item_id"],
            "msg": "String should have at least 2 characters",
            "type": "string_too_short",
            "ctx": {"min_length": 2},
        }
    ]
}


response_maximum_3 = {
    "detail": [
        {
            "loc": ["path", "item_id"],
            "msg": "String should have at most 3 characters",
            "type": "string_too_long",
            "ctx": {"max_length": 3},
        }
    ]
}


response_greater_than_3 = {
    "detail": [
        {
            "loc": ["path", "item_id"],
            "msg": "Input should be greater than 3",
            "type": "greater_than",
            "ctx": {"gt": 3.0},
        }
    ]
}


response_greater_than_0 = {
    "detail": [
        {
            "loc": ["path", "item_id"],
            "msg": "Input should be greater than 0",
            "type": "greater_than",
            "ctx": {"gt": 0.0},
        }
    ]
}


response_greater_than_1 = {
    "detail": [
        {
            "loc": ["path", "item_id"],
            "msg": "Input should be greater than 1",
            "type": "greater_than",
            "ctx": {"gt": 1.0},
        }
    ]
}


response_greater_than_equal_3 = {
    "detail": [
        {
            "loc": ["path", "item_id"],
            "msg": "Input should be greater than or equal to 3",
            "type": "greater_than_equal",
            "ctx": {"ge": 3.0},
        }
    ]
}


response_less_than_3 = {
    "detail": [
        {
            "loc": ["path", "item_id"],
            "msg": "Input should be less than 3",
            "type": "less_than",
            "ctx": {"lt": 3.0},
        }
    ]
}


response_less_than_0 = {
    "detail": [
        {
            "loc": ["path", "item_id"],
            "msg": "Input should be less than 0",
            "type": "less_than",
            "ctx": {"lt": 0.0},
        }
    ]
}


response_less_than_equal_3 = {
    "detail": [
        {
            "loc": ["path", "item_id"],
            "msg": "Input should be less than or equal to 3",
            "type": "less_than_equal",
            "ctx": {"le": 3.0},
        }
    ]
}


valid_foobar_str = {"item_id": "foobar"}
valid_42_str = {"item_id": "42"}
valid_true_str = {"item_id": "True"}
valid_true_bool = {"item_id": True}
valid_false_bool = {"item_id": False}
valid_42_int = {"item_id": 42}
valid_42_5_float = {"item_id": 42.5}


def _valid(x):
    return {"item_id": x}


not_found = {"message": "Not Found"}


@pytest.mark.parametrize(
    "url,expected_status,expected_response",
    [
        ("/path/foobar", 200, valid_foobar_str),
        ("/path/str/foobar", 200, valid_foobar_str),
        ("/path/str/42", 200, valid_42_str),
        ("/path/str/True", 200, valid_true_str),
        ("/path/int/foobar", 400, response_not_valid_int),
        ("/path/int/True", 400, response_not_valid_int),
        ("/path/int/42", 200, valid_42_int),
        ("/path/int/42.5", 400, response_not_valid_int),
        ("/path/float/foobar", 400, response_not_valid_float),
        ("/path/float/True", 400, response_not_valid_float),
        ("/path/float/42", 200, _valid(42.0)),
        ("/path/float/42.5", 200, valid_42_5_float),
        ("/path/bool/foobar", 400, response_not_valid_bool),
        ("/path/bool/True", 200, valid_true_bool),
        ("/path/bool/42", 400, response_not_valid_bool),
        ("/path/bool/42.5", 400, response_not_valid_bool),
        ("/path/bool/1", 200, valid_true_bool),
        ("/path/bool/0", 200, valid_false_bool),
        ("/path/bool/true", 200, valid_true_bool),
        ("/path/bool/False", 200, valid_false_bool),
        ("/path/bool/false", 200, valid_false_bool),
        ("/path/param/foo", 200, _valid("foo")),
        ("/path/param-required/foo", 200, _valid("foo")),
        ("/path/param-minlength/foo", 200, _valid("foo")),
        ("/path/param-minlength/fo", 400, response_at_least_3),
        ("/path/param-maxlength/foo", 200, _valid("foo")),
        ("/path/param-maxlength/foobar", 400, response_maximum_3),
        ("/path/param-min_maxlength/foo", 200, _valid("foo")),
        ("/path/param-min_maxlength/foobar", 400, response_maximum_3),
        ("/path/param-min_maxlength/f", 400, response_at_least_2),
        ("/path/param-gt/42", 200, valid_42_int),
        ("/path/param-gt/2", 400, response_greater_than_3),
        ("/path/param-gt0/0.05", 200, _valid(0.05)),
        ("/path/param-gt0/0", 400, response_greater_than_0),
        ("/path/param-ge/42", 200, valid_42_int),
        ("/path/param-ge/3", 200, _valid(3)),
        ("/path/param-ge/2", 400, response_greater_than_equal_3),
        ("/path/param-lt/42", 400, response_less_than_3),
        ("/path/param-lt/2", 200, _valid(2)),
        ("/path/param-lt0/-1", 200, _valid(-1)),
        ("/path/param-lt0/0", 400, response_less_than_0),
        ("/path/param-le/42", 400, response_less_than_equal_3),
        ("/path/param-le/3", 200, _valid(3)),
        ("/path/param-le/2", 200, _valid(2)),
        ("/path/param-lt-gt/2", 200, _valid(2)),
        ("/path/param-lt-gt/4", 400, response_less_than_3),
        ("/path/param-lt-gt/0", 400, response_greater_than_1),
        ("/path/param-le-ge/2", 200, _valid(2)),
        ("/path/param-le-ge/1", 200, _valid(1)),
        ("/path/param-le-ge/3", 200, _valid(3)),
        ("/path/param-le-ge/4", 400, response_less_than_equal_3),
        ("/path/param-lt-int/2", 200, _valid(2)),
        ("/path/param-lt-int/42", 400, response_less_than_3),
        ("/path/param-lt-int/2.7", 400, response_not_valid_int),
        ("/path/param-gt-int/42", 200, valid_42_int),
        ("/path/param-gt-int/2", 400, response_greater_than_3),
        ("/path/param-gt-int/2.7", 400, response_not_valid_int),
        ("/path/param-le-int/42", 400, response_less_than_equal_3),
        ("/path/param-le-int/3", 200, _valid(3)),
        ("/path/param-le-int/2", 200, _valid(2)),
        ("/path/param-le-int/2.7", 400, response_not_valid_int),
        ("/path/param-ge-int/42", 200, valid_42_int),
        ("/path/param-ge-int/3", 200, _valid(3)),
        ("/path/param-ge-int/2", 400, response_greater_than_equal_3),
        ("/path/param-ge-int/2.7", 400, response_not_valid_int),
        ("/path/param-lt-gt-int/2", 200, _valid(2)),
        ("/path/param-lt-gt-int/4", 400, response_less_than_3),
        ("/path/param-lt-gt-int/0", 400, response_greater_than_1),
        ("/path/param-lt-gt-int/2.7", 400, response_not_valid_int),
        ("/path/param-le-ge-int/2", 200, _valid(2)),
        ("/path/param-le-ge-int/1", 200, _valid(1)),
        ("/path/param-le-ge-int/3", 200, _valid(3)),
        ("/path/param-le-ge-int/4", 400, response_less_than_equal_3),
        ("/path/param-le-ge-int/2.7", 400, response_not_valid_int),
    ],
)
def test_get_path(client, url, expected_status, expected_response):
    """Test Path Operation with GET method.


    TODO: Allow Configuration of the returned Status Code.
    """
    response = client.get(url)
    assert response.status_code == expected_status
    assert response.json == expected_response


@pytest.mark.parametrize(
    "url,expected_status,expected_response",
    [
        ("/path/with_converter/foobar", 200, valid_foobar_str),
        ("/path/with_converter/str/foobar", 200, valid_foobar_str),
        ("/path/with_converter/str/42", 200, valid_42_str),
        ("/path/with_converter/str/True", 200, _valid("True")),
        ("/path/with_converter/int/42", 200, valid_42_int),
        ("/path/with_converter/float/42.5", 200, valid_42_5_float),
        ("/path/with_converter/param/foo", 200, _valid("foo")),
        ("/path/with_converter/param-required/foo", 200, _valid("foo")),
        ("/path/with_converter/param-minlength/foo", 200, _valid("foo")),
        ("/path/with_converter/param-minlength/fo", 400, response_at_least_3),
        ("/path/with_converter/param-maxlength/foo", 200, _valid("foo")),
        ("/path/with_converter/param-maxlength/foobar", 400, response_maximum_3),
        ("/path/with_converter/param-min_maxlength/foo", 200, _valid("foo")),
        ("/path/with_converter/param-min_maxlength/foobar", 400, response_maximum_3),
        ("/path/with_converter/param-min_maxlength/f", 400, response_at_least_2),
        ("/path/with_converter/param-gt/42.0", 200, _valid(42.0)),
        ("/path/with_converter/param-gt/2.0", 400, response_greater_than_3),
        ("/path/with_converter/param-gt0/0.05", 200, _valid(0.05)),
        ("/path/with_converter/param-gt0/0.0", 400, response_greater_than_0),
        ("/path/with_converter/param-ge/42.0", 200, _valid(42.0)),
        ("/path/with_converter/param-ge/3.0", 200, _valid(3.0)),
        ("/path/with_converter/param-ge/2.0", 400, response_greater_than_equal_3),
        ("/path/with_converter/param-lt/42.0", 400, response_less_than_3),
        ("/path/with_converter/param-lt/2.0", 200, _valid(2.0)),
        ("/path/with_converter/param-lt0/-1.0", 200, _valid(-1)),
        ("/path/with_converter/param-lt0/0.0", 400, response_less_than_0),
        ("/path/with_converter/param-le/42.0", 400, response_less_than_equal_3),
        ("/path/with_converter/param-le/3.0", 200, _valid(3)),
        ("/path/with_converter/param-le/2.0", 200, _valid(2)),
        ("/path/with_converter/param-lt-gt/2.0", 200, _valid(2)),
        ("/path/with_converter/param-lt-gt/4.0", 400, response_less_than_3),
        ("/path/with_converter/param-lt-gt/0.0", 400, response_greater_than_1),
        ("/path/with_converter/param-le-ge/2.0", 200, _valid(2)),
        ("/path/with_converter/param-le-ge/1.0", 200, _valid(1)),
        ("/path/with_converter/param-le-ge/3.0", 200, _valid(3)),
        ("/path/with_converter/param-le-ge/4.0", 400, response_less_than_equal_3),
        ("/path/with_converter/param-lt-int/2", 200, _valid(2)),
        ("/path/with_converter/param-lt-int/42", 400, response_less_than_3),
        ("/path/with_converter/param-gt-int/42", 200, _valid(42)),
        ("/path/with_converter/param-gt-int/2", 400, response_greater_than_3),
        ("/path/with_converter/param-le-int/42", 400, response_less_than_equal_3),
        ("/path/with_converter/param-le-int/3", 200, _valid(3)),
        ("/path/with_converter/param-le-int/2", 200, _valid(2)),
        ("/path/with_converter/param-ge-int/42", 200, valid_42_int),
        ("/path/with_converter/param-ge-int/3", 200, _valid(3)),
        ("/path/with_converter/param-ge-int/2", 400, response_greater_than_equal_3),
        ("/path/with_converter/param-lt-gt-int/2", 200, _valid(2)),
        ("/path/with_converter/param-lt-gt-int/4", 400, response_less_than_3),
        ("/path/with_converter/param-lt-gt-int/0", 400, response_greater_than_1),
        ("/path/with_converter/param-le-ge-int/2", 200, _valid(2)),
        ("/path/with_converter/param-le-ge-int/1", 200, _valid(1)),
        ("/path/with_converter/param-le-ge-int/3", 200, _valid(3)),
        ("/path/with_converter/param-le-ge-int/4", 400, response_less_than_equal_3),
    ],
)
def test_get_path_with_converter(client, url, expected_status, expected_response):
    """Test Path Operation with GET method.


    TODO: Allow Configuration of the returned Status Code.
    """
    response = client.get(url)
    assert response.json == expected_response
    assert response.status_code == expected_status


@pytest.mark.parametrize(
    "url,expected_status,expected_response",
    [
        ("/path/with_converter/int/foobar", 404, not_found),
        ("/path/with_converter/int/True", 404, not_found),
        ("/path/with_converter/int/42.5", 404, not_found),
        ("/path/with_converter/float/foobar", 404, not_found),
        ("/path/with_converter/float/True", 404, not_found),
        ("/path/with_converter/float/42", 404, not_found),
        ("/path/with_converter/param-le-int/2.7", 404, not_found),
        ("/path/with_converter/param-ge-int/2.7", 404, not_found),
        ("/path/with_converter/param-le-ge-int/2.7", 404, not_found),
    ],
)
def test_path_converter_error_override_jeroboam_validation(
    client, url, expected_status, expected_response
):
    """Test Url Converter Overides PathParams Validation.

    GIVEN a Path Parameter with both Path (Jeroboam) and Converter (Flask) validation
    WHEN the url is called with a value that does not match the converter
    THEN the converter error is returned (404), not the Jeroboam one (400)
    """
    response = client.get(url)
    assert response.json == expected_response
    assert response.status_code == expected_status



"""Explicit behavioral tests for InboundModel and OutboundModel.

These tests pin the camelCase ↔ snake_case conversion behavior driven by
Pydantic's alias_generator and allow_population_by_field_name / populate_by_name.

Both of these config keys change names in Pydantic v2:
  - allow_population_by_field_name → populate_by_name
  - alias_generator stays, but must return str (same as before)
  - Config inner class → model_config = ConfigDict(...)

If these tests break after the migration, the InboundModel/OutboundModel
config was not migrated correctly.
"""

import pytest
from flask.testing import FlaskClient

from flask_jeroboam import Jeroboam
from flask_jeroboam.models import InboundModel, OutboundModel

# ---------------------------------------------------------------------------
# InboundModel: camelCase query param acceptance
# The /query/optional_model route uses OptionalModelIn(InboundModel) which has
# a per_page: int | None field. InboundModel's alias_generator creates the
# "perPage" alias, and allow_population_by_field_name allows both to work.
# ---------------------------------------------------------------------------


@pytest.mark.parametrize(
    "query_string,expected_json",
    [
        # camelCase alias works
        ("?page=2&perPage=5", {"page": 2, "per_page": 5}),
        # snake_case field name also works (allow_population_by_field_name)
        ("?page=2&per_page=5", {"page": 2, "per_page": 5}),
        # partial: only page provided, per_page is None
        ("?page=3", {"page": 3, "per_page": None}),
        # empty: both None
        ("", {"page": None, "per_page": None}),
    ],
)
def test_inbound_model_accepts_both_camel_and_snake_case(
    client: FlaskClient, query_string, expected_json
):
    """GIVEN an endpoint with an InboundModel parameter
    WHEN queried with camelCase or snake_case query params
    THEN both are accepted and stored internally as snake_case.

    This validates that:
    - alias_generator correctly generates camelCase aliases
    - allow_population_by_field_name / populate_by_name allows the field name too
    """
    response = client.get(f"/query/optional_model{query_string}")
    assert response.status_code == 200
    assert response.json == expected_json


# ---------------------------------------------------------------------------
# OutboundModel: snake_case serialized as camelCase
# The /return_type/dict route uses SimpleModelOut(OutboundModel) which has
# a total_count field. OutboundModel serializes it as totalCount.
# ---------------------------------------------------------------------------


def test_outbound_model_serializes_snake_case_to_camel_case(client: FlaskClient):
    """GIVEN an endpoint with an OutboundModel response_model
    WHEN the view returns snake_case field names
    THEN the JSON response uses camelCase.

    This validates that OutboundModel's json serialization is configured correctly.
    Pydantic v1: json_dumps = json_dumps_to_camel_case (custom Config option)
    Pydantic v2: requires different approach (alias_generator + by_alias=True)
    """
    response = client.get("/return_type/dict")
    assert response.status_code == 200
    assert "totalCount" in response.json
    assert "total_count" not in response.json
    assert response.json["totalCount"] == 10


# ---------------------------------------------------------------------------
# Unit-level: InboundModel and OutboundModel class behaviour
# ---------------------------------------------------------------------------


def test_inbound_model_parses_camel_case_directly():
    """GIVEN an InboundModel subclass with a snake_case field
    WHEN instantiated with a camelCase key
    THEN the value is stored under the snake_case field name.
    """

    class Pagination(InboundModel):
        per_page: int = 10

    instance = Pagination(perPage=25)
    assert instance.per_page == 25


def test_inbound_model_parses_snake_case_directly():
    """GIVEN an InboundModel subclass with a snake_case field
    WHEN instantiated with the snake_case key directly
    THEN the value is accepted (allow_population_by_field_name / populate_by_name).
    """

    class Pagination(InboundModel):
        per_page: int = 10

    instance = Pagination(per_page=30)
    assert instance.per_page == 30


def test_outbound_model_serializes_to_camel_case_directly():
    """GIVEN an OutboundModel subclass with a snake_case field
    WHEN serialized to JSON
    THEN the key is camelCase in the output.
    """

    class Stats(OutboundModel):
        total_count: int

    instance = Stats(total_count=42)
    import json

    data = json.loads(instance.model_dump_json())
    assert "totalCount" in data
    assert data["totalCount"] == 42


def test_inbound_model_validation_error_reports_snake_case_field_name(
    one_shot_app: Jeroboam,
):
    """GIVEN an endpoint with a required InboundModel field
    WHEN the request omits that field entirely
    THEN the error loc uses the Python field name, not the camelCase alias.

    This ensures validation errors are readable in Python terms.
    """

    class StrictPagination(InboundModel):
        per_page: int  # required, no default

    @one_shot_app.get("/test_strict_inbound")
    def strict_endpoint(pagination: StrictPagination):
        return {}

    client = one_shot_app.test_client()
    response = client.get("/test_strict_inbound")
    assert response.status_code == 400
    locs = [e.get("loc", []) for e in response.json["detail"]]
    # The field name in the error should be per_page (or perPage alias)
    # — but crucially the error must be present
    assert any("per_page" in str(loc) or "perPage" in str(loc) for loc in locs)



from collections.abc import Callable
from typing import Any

from flask_jeroboam.typing import JeroboamRouteCallable

class JeroboamScaffoldOverRide:
    def route(
        self, rule: str, **options: Any
    ) -> Callable[[JeroboamRouteCallable], JeroboamRouteCallable]: ...
    def get(
        self, rule: str, **options: Any
    ) -> Callable[[JeroboamRouteCallable], JeroboamRouteCallable]: ...
    def post(
        self, rule: str, **options: Any
    ) -> Callable[[JeroboamRouteCallable], JeroboamRouteCallable]: ...
    def put(
        self, rule: str, **options: Any
    ) -> Callable[[JeroboamRouteCallable], JeroboamRouteCallable]: ...
    def delete(
        self, rule: str, **options: Any
    ) -> Callable[[JeroboamRouteCallable], JeroboamRouteCallable]: ...
    def patch(
        self, rule: str, **options: Any
    ) -> Callable[[JeroboamRouteCallable], JeroboamRouteCallable]: ...
    def _method_route(
        self,
        method: str,
        rule: str,
        options: dict,
    ) -> Callable[[JeroboamRouteCallable], JeroboamRouteCallable]: ...



"""Configuration for Generating Docs."""

# import packaging.version
from pallets_sphinx_themes import (
    ProjectLink,  # type: ignore
    get_version,  # type: ignore
)

# Project --------------------------------------------------------------

project = "Flask-Jeroboam"
copyright = "2022, Jean-Christophe Bianic"
author = "Jean-Christophe Bianic"
release, version = get_version("Flask-Jeroboam")

# General --------------------------------------------------------------

master_doc = "index"
extensions = [
    "pallets_sphinx_themes",
    "sphinx.ext.autodoc",
    "sphinx.ext.napoleon",
    "sphinx.ext.autosectionlabel",
    "sphinx_click",
    "myst_parser",
    "sphinx_multiversion",
    "sphinx_tabs.tabs",
    "sphinx_copybutton",
]
autodoc_typehints = "description"

smv_tag_whitelist = r"^v\d+\.\d+.*$|latest"
smv_branch_whitelist = r"^(?main).*$"
smv_released_pattern = r"v.*"
smv_remote_whitelist = r"^(origin)$"

# HTML -----------------------------------------------------------------

html_theme = "flask"
html_theme_options = {"index_sidebar_logo": False}
html_context = {
    "project_links": [
        ProjectLink("PyPI Releases", "https://pypi.org/project/flask-jeroboam/"),
        ProjectLink("Source Code", "https://github.com/jcbianic/flask-jeroboam"),
        ProjectLink(
            "Issue Tracker", "https://github.com/jcbianic/flask-jeroboam/issues/"
        ),
    ]
}
# html_sidebars = {
#     "index": ["project.html", "localtoc.html", "searchbox.html", "ethicalads.html"],
#     "**": ["localtoc.html", "relations.html", "searchbox.html", "ethicalads.html"],
# }
# singlehtml_sidebars = {"index": ["project.html", "localtoc.html", "ethicalads.html"]}
html_static_path = ["_static"]
templates_path = ["_templates"]
html_favicon = "_static/img/jeroboam_icon.png"
html_logo = "_static/img/jeroboam_logo_with_text.png"
html_title = f"Flask-Jeroboam Documentation ({version})"
html_show_sourcelink = False
html_css_files = ["custom.css"]


# CopyButton -----------------------------------------------------------------

copybutton_prompt_text = r">>> |\.\.\. |\$ "
copybutton_prompt_is_regexp = True
copybutton_only_copy_prompt_lines = True


# MISCELLANEOUS -------------------------------------------------------------

autosectionlabel_prefix_document = True



"""Fictional CRUD operations."""

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from docs_src.readme.readme01 import GenericPagination

wines: list[dict[str, str]] = [
    {
        "appellation": "Margaux",
        "domain": "Château Magaux",
        "cuvee": "Pavillon Rouge",
        "color": "Rouge",
    },
    {
        "appellation": "Meursault",
        "domain": "Domaine Comte Armand ",
        "cuvee": "Meursault",
        "color": "Blanc",
    },
    {
        "appellation": "Champagne",
        "domain": "Billecart-Salmon",
        "cuvee": "Brut - Blanc de Blancs",
        "color": "Blanc",
    },
    {
        "appellation": "Champagne",
        "domain": "Krug",
        "cuvee": "Grande Cuvée - 170ème Edition",
        "color": "Blanc",
    },
    {
        "appellation": "Champagne",
        "domain": "Maison Taittinger",
        "cuvee": "Grand Cru - Brut - Prélude",
        "color": "Blanc",
    },
]


def get_wines(pagination: "GenericPagination", search: str | None) -> list[dict]:
    """Get wines from a list."""
    if search:
        wines_filtered = [
            wine
            for wine in wines
            if search.lower() in f"{wine['appellation']}-{wine['cuvee']}".lower()
        ]
    else:
        wines_filtered = wines
    return wines_filtered[pagination.offset : pagination.offset + pagination.per_page]



"""Response with custom status code."""

from pydantic import BaseModel

from flask_jeroboam import Jeroboam

app = Jeroboam("Jeroboam Outbound Features App")
app.init_app()


class Task(BaseModel):
    id: int
    title: str


@app.post("/tasks", response_model=Task, status_code=201)
def create_task(title: str):
    return {"id": 1, "title": title}


if __name__ == "__main__":
    app.run(host="localhost", port=5000)



"""The Blueprint for the OpenAPI UI."""

from typing import TYPE_CHECKING

from flask import render_template

from flask_jeroboam.blueprint import Blueprint
from flask_jeroboam.openapi.models.openapi import OpenAPI
from flask_jeroboam.openapi.models.ui_context import SwaggerContextOut
from flask_jeroboam.responses import HTMLResponse, JSONResponse
from flask_jeroboam.wrapper import current_app

if TYPE_CHECKING:  # pragma: no cover
    from flask_jeroboam.jeroboam import Jeroboam

router = Blueprint(
    "openapi_docs",
    __name__,
    template_folder="templates",
    static_folder="different_name",
    include_in_openapi=False,
)


@router.get("/docs", include_in_openapi=False)
def get_swagger_html():
    """Serving the Swagger UI HTML."""
    context = SwaggerContextOut(
        title=current_app.config.get("JEROBOAM_TITLE", "Jeroboam App")
    )
    return HTMLResponse(render_template("swagger-ui.jinja", **context.model_dump()))


@router.get("/openapi.json", response_model=OpenAPI, include_in_openapi=False)
def get_openapi_json():
    """Serving OpenAPI JSON."""
    return JSONResponse(
        current_app.openapi.model_dump_json(exclude_none=True, by_alias=True)
    )


def register_open_api_blueprint(app: "Jeroboam") -> None:
    """Register the OpenAPI Blueprint."""
    app.register_blueprint(router)



"""Test the FileParam.

The corresponding endpoints are defined in the test_app.apps.file.py module.
"""

from io import BytesIO

from flask.testing import FlaskClient


def test_valid_payload_in_files_is_injected(client: FlaskClient):
    """GIVEN a POST endpoint with FileParam
    WHEN hit with a valid files payload
    THEN the parsed input is injected into the view function.
    """
    data = {"file": (BytesIO(b"Hello World !!"), "hello.txt"), "type": "file"}

    response = client.post(
        "/file",
        data=data,
        headers={
            "enctype": "multipart/form-data",
        },
    )

    assert response.json == {"file_content": "b'Hello World !!'"}
    assert response.status_code == 201



"""A Test Blueprint for testing OpenApi Specials.

The corresponding test can be found in tests/test_openapi/*
"""

from flask_jeroboam import Blueprint
from flask_jeroboam.view_arguments.functions import Query

router = Blueprint("openapi_args_router", __name__, tags=["OpenApi"])


@router.get("/query/dont_include_in_schema")
def read_argument_not_included_in_schema(
    not_included: int = Query(include_in_schema=False),
):
    return {"not_included": not_included}


@router.get(
    "/query/openapi_extra",
    openapi_extra={"example": {"testA": 1}, "examples": [{"test": 1}]},
)
def read_openapi_extra(
    with_openapi_extra: int = Query(
        include_in_schema=False,
    ),
):
    return {"with_openapi_extra": with_openapi_extra}


@router.get("/openapi/no_body_status_code", status_code=205)
def get_that_returns_a_empty_body(payload: str):
    return None


@router.get("/openapi/no_response_model", response_model=None)
def get_without_a_reponse_model(payload: str):
    return None



"""Defining Fixtures for the Test Suite."""

import pytest
from flask.testing import FlaskClient

from flask_jeroboam import Jeroboam
from tests.app_test.app import create_test_app


@pytest.fixture(scope="session")
def app() -> Jeroboam:
    """The Jeroboam Test App."""
    return create_test_app()


@pytest.fixture(scope="function")
def one_shot_app() -> Jeroboam:
    """The Jeroboam Test App."""
    return create_test_app()


@pytest.fixture(scope="function")
def one_shot_client(one_shot_app: Jeroboam) -> FlaskClient:
    """The Jeroboam Test App."""
    return one_shot_app.test_client()


@pytest.fixture
def app_ctx(app: Jeroboam):
    """Application Context from the Test App."""
    with app.app_context() as ctx:
        yield ctx


@pytest.fixture
def request_context(app: Jeroboam):
    """Request Context from the Test App."""
    with app.test_request_context() as ctx:
        yield ctx


@pytest.fixture
def client(app: Jeroboam) -> FlaskClient:
    """Test Client from the Test App."""
    return app.test_client()



"""Testing for status code configuration.

Endpoints are defined in the app_test.apps.outbound.py module.
"""

import warnings
from unittest.mock import patch

import pytest
from flask.testing import FlaskClient

from flask_jeroboam.jeroboam import Jeroboam
from tests.app_test.models.outbound import SimpleModelOut

valid_outbound_data = {"items": ["Apple", "Banana"], "total_count": 10}
valid_response_body = {"items": ["Apple", "Banana"], "totalCount": 10}
unsorted_reponse_body = {"total_count": 10, "items": ["Apple", "Banana"]}


def test_endpoint_with_put(
    client: FlaskClient,
):
    """GIVEN an endpoint registered wit put
    WHEN hit with a put request
    THEN it responds with a 201 status code
    """
    response = client.put("/verb/put/without_explicit_status_code")
    assert response.status_code == 201
    assert response.json == valid_response_body


def test_endpoint_with_patch(
    client: FlaskClient,
):
    """GIVEN an endpoint registered wit put
    WHEN hit with a patch request
    THEN it responds with a 201 status code
    """
    response = client.patch("/verb/patch/without_explicit_status_code")
    assert response.status_code == 200
    assert response.json == valid_response_body


@pytest.mark.parametrize("variant", ["as_returned", "as_configured"])
def test_status_code_204_has_no_body(
    variant: str,
    client: FlaskClient,
):
    """GIVEN a 204 Status Code
    WHEN building the response
    THEN the response has no body
    """
    response = client.get(f"/status_code/204_has_no_body/{variant}")
    assert response.status_code == 204
    assert response.data == b""


@patch("flask_jeroboam._outboundhandler.METHODS_DEFAULT_STATUS_CODE", {"POST": 201})
def test_exotic_http_verb_raise_a_warning_when_no_status_code_is_set(
    one_shot_app: Jeroboam,
    one_shot_client: FlaskClient,
):
    """GIVEN an endpoint with an exotic HTTP verb and no status_code defined
    WHEN registered
    THEN a User Warning is raised
    """
    with pytest.warns(UserWarning):

        @one_shot_app.get(
            "/exotic_http_verb_raise_a_warning",
            response_model=SimpleModelOut,
        )
        def exotic_http_verb():
            return valid_outbound_data

    response = one_shot_client.get("/exotic_http_verb_raise_a_warning")

    assert response.status_code == 200


@patch("flask_jeroboam._outboundhandler.METHODS_DEFAULT_STATUS_CODE", {"POST": 201})
def test_exotic_http_verb_dont_raise_a_warning_when_status_code_is_set(
    one_shot_app: Jeroboam,
    one_shot_client: FlaskClient,
):
    """GIVEN an endpoint with an exotic HTTP verb and no status_code defined
    WHEN registered
    THEN no User Warning is raised
    """
    with warnings.catch_warnings():
        warnings.simplefilter("error")

        @one_shot_app.get(
            "/exotic_http_verb_dont_raise_a_warning",
            response_model=SimpleModelOut,
            status_code=200,
        )
        def exotic_http_verb():
            return valid_outbound_data

    response = one_shot_client.get("/exotic_http_verb_dont_raise_a_warning")

    assert response.status_code == 200



"""Code Sample for OpenAPI Features Documentation."""

from pydantic import BaseModel

from flask_jeroboam import Blueprint, Jeroboam

app = Jeroboam("Jeroboam Inbound Features App")
app.init_app()


class TaskIn(BaseModel):
    name: str
    description: str | None = None


class TaskOut(TaskIn):
    id: int


@app.get("/health")
def get_health():
    return {"status": "Ok"}


blueprint = Blueprint("tasks", __name__, tags=["tasks"])


@blueprint.get("/tasks", response_model=list[TaskOut])
def get_tasks(page: int = 1, per_page: int = 10, search: str | None = None):
    return [{"id": 1, "name": "Task 1"}, {"id": 2, "name": "Task 2"}]


@blueprint.get("/tasks/<int:item_id>", response_model=TaskOut)
def get_task(item_id: int):
    return {"id": 1, "name": "Task 1"}


@blueprint.put("/tasks", status_code=202, tags=["sensitive"])
def create_task(task: TaskIn):
    return {}


@blueprint.post("/tasks/<int:item_id>", tags=["sensitive"])
def edit_task(item_id: int, task: TaskIn):
    return {}


app.register_blueprint(blueprint)

if __name__ == "__main__":
    app.run(host="localhost", port=5000)



"""A Test Blueprint for testing File Params.

The corresponding test can be found in tests/test_inbound/test_file
"""

from flask_jeroboam import Blueprint, File
from flask_jeroboam.datastructures import UploadFile

router = Blueprint("file_params_router", __name__, tags=["File"])


@router.post("/file")
def ping(file: UploadFile = File()):
    return {"file_content": str(file.read())}



"""Testing Outbound Handler Use Cases.

The endpoints are defined in the app_test.apps.outbound.py module.
"""

from typing import Any

import pytest
from flask.testing import FlaskClient

from flask_jeroboam.jeroboam import Jeroboam
from tests.app_test.models.outbound import SimpleModelOut

valid_outbound_data = {"items": ["Apple", "Banana"], "total_count": 10}
valid_response_body = {"items": ["Apple", "Banana"], "totalCount": 10}
unsorted_reponse_body = {"total_count": 10, "items": ["Apple", "Banana"]}


@pytest.mark.parametrize(
    "url", ["/methods/explicit_options", "/methods/explicit_options_and_head"]
)
def test_register_route_with_additionnal_secondary_verb(
    url: str,
    client: FlaskClient,
):
    """GIVEN an endpoint with secondary methods defined
    WHEN configured with secondary verbs (OPTIONS, HEAD)
    THEN it register and keeps the main verb
    """
    response = client.get(url)

    assert response.json == valid_response_body
    assert response.status_code == 200


def test_register_route_with_two_main_verb_raise_a_warning(
    one_shot_app: Jeroboam,
    one_shot_client: FlaskClient,
):
    """GIVEN an endpoint configured with two main http verb
    WHEN it is registered
    THEN it raises a warning and keep the first one
    """
    with pytest.warns(UserWarning):

        @one_shot_app.route(
            "/endpoint_with_two_main_verb",
            methods=["GET", "POST"],
            response_model=SimpleModelOut,
        )
        def with_two_main_verb():
            return valid_outbound_data

    response = one_shot_client.get("/endpoint_with_two_main_verb")
    assert response.json == valid_response_body


def test_register_route_with_method_route_and_methods_option_raise_a_exception(
    one_shot_app: Jeroboam,
):
    """GIVEN an endpoint configured with the method_route
    WHEN it is registered with the methods option
    THEN it raises an Exception
    """
    with pytest.raises(TypeError):

        @one_shot_app.get(
            "/route_method_and_methods_option",
            methods=["GET"],
            response_model=SimpleModelOut,
        )
        def route_method_and_methods_option():
            return valid_outbound_data


def test_endpoint_without_response_model(
    client: FlaskClient,
):
    """GIVEN an endpoint with no response_model configured or return annotation
    WHEN hit
    THEN it behaves like a regular Flask endpoint
    """
    response = client.get("/response_model/no_response_model")
    assert response.status_code == 200
    assert response.data == b"Don't have a response model"


def test_endpoint_with_valid_return_annocation(
    client: FlaskClient,
):
    """GIVEN an endpoint without response_model but with a valid return annotation
    WHEN registered
    THEN the annotation is stored as response_model
    """
    response = client.get("/response_model/infered_from_return_annotation")
    assert response.status_code == 200
    assert response.json == valid_response_body


def test_invalid_response_model_raise_type_error_at_registration(
    one_shot_app: Jeroboam,
):
    """GIVEN an endpoint with invalid response_model
    WHEN registered
    THEN it raises a TypeError
    """
    with pytest.raises(TypeError):

        @one_shot_app.get("/invalid_return_annotation_and_no_response_model")
        def invalid_return_annotation() -> dict:
            return valid_outbound_data

    with pytest.raises(TypeError):

        @one_shot_app.get("/invalid_response_model", response_model=dict)
        def invalid_configuration():
            return valid_outbound_data


def test_configured_response_model_take_prescedence_over_return_annotation(
    one_shot_client: FlaskClient,
):
    """GIVEN an endpoint without a configured response_model and a return annotation
    WHEN registered
    THEN the configrued response_model take prescedence over the return annotation
    """
    response = one_shot_client.get("/response_model/configuration_over_inference")
    assert response.status_code == 200
    assert response.json == valid_response_body


def test_endpoint_can_turn_off_return_annocation(
    client: FlaskClient,
):
    """GIVEN an endpoint with a valid return annotation
    WHEN response_model is configured to be None
    THEN the response_model registration is turned off
    """
    with pytest.raises(TypeError):
        client.get("/response_model/turned_off")


@pytest.mark.parametrize(
    "type_,expected_response",
    [
        ("dict", valid_response_body),
        ("list", [valid_response_body, valid_response_body]),
        ("base_model", valid_response_body),
        ("response", valid_response_body),
        ("dataclass", valid_response_body),
    ],
)
def test_view_function_with_response_model_return_type(
    type_: str,
    expected_response: Any,
    client: FlaskClient,
):
    """GIVEN an endpoint with a response_model defined and dict return value
    WHEN hit
    THEN it serializes the response using the response_model, applying field aliases
    """
    response = client.get(f"/return_type/{type_}")

    assert response.status_code == 200
    assert response.json == expected_response


def test_wrong_dict_being_sent(
    one_shot_app: Jeroboam,
    one_shot_client: FlaskClient,
):
    """GIVEN an endpoint with a response_model defined and a dict return value
    WHEN hit and the return value is not valid
    THEN it raises a InternalServerError, 500
    """

    @one_shot_app.get("/invalid_return_value", response_model=SimpleModelOut)
    def ping():
        return {"total_count": "not_valid", "items": ["Apple", "Banana"]}

    response = one_shot_client.get("/invalid_return_value")

    assert response.status_code == 500
    assert response.data.startswith(b"InternalServerError")


@pytest.mark.parametrize(
    "shape,status_code,headers",
    [
        ("with_headers", 200, {"X-Test": "Test"}),
        ("with_status_code", 218, {}),
        ("with_headers_and_status_code", 218, {"X-Test": "Test"}),
    ],
)
def test_view_function_tuple_return_shape(
    shape: str,
    status_code: int,
    headers: dict[str, str],
    client: FlaskClient,
):
    """GIVEN an endpoint with a response_model defined and a dict return value
    WHEN hit and the return value is not valid
    THEN it raises a InternalServerError, 500
    """
    response = client.get(f"/return_shape/{shape}")
    assert response.status_code == status_code
    assert response.headers.get("X-Test", "Empty") == headers.get("X-Test", "Empty")


def test_wrong_tuple_length_raise_error(
    client: FlaskClient,
):
    """GIVEN a viewfunction with the wrongly shaped tuple (>3)
    WHEN hit
    THEN it raises a TypeError (as in Flask) and return a code 500
    """
    with pytest.raises(TypeError):
        respone = client.get("/return_shape/wrong_tuple_length")

        assert respone.status_code == 500


def test_content_raise_an_error_if_anything_else(
    client: FlaskClient,
):
    """GIVEN an endpoint with an exotic HTTP verb and no status_code defined
    WHEN registered
    THEN a Value Warning is raised
    """
    with pytest.raises(ValueError):
        response = client.get("/return_type/not_valid")

        assert response.status_code == 500


def test_reponse_model_filters_outbound_data_even_when_subclassing(
    client: FlaskClient,
):
    """GIVEN an endpoint with an exotic HTTP verb and no status_code defined
    WHEN registered
    THEN a User Warning is raised
    """
    response = client.post(
        "/sensitive_data",
        json={"username": "test", "password": "test"},
    )

    assert response.json == {"username": "test"}
    assert response.status_code == 201



"""A Test Blueprint for testing mixed parameter locations.

Tests that path, query and body parameters are all resolved correctly
when used together in the same endpoint.
The corresponding tests are in tests/test_inbound_handler/test_mixed_parameters.py
"""

from pydantic import BaseModel

from flask_jeroboam import Blueprint, Query
from flask_jeroboam.view_arguments.functions import Body

router = Blueprint("mixed_router", __name__, tags=["Mixed"])


class ItemIn(BaseModel):
    name: str
    price: float


class ItemOut(BaseModel):
    id: int
    name: str
    price: float
    q: str | None = None


@router.get("/mixed/<int:item_id>")
def get_item(item_id: int, q: str | None = None):
    """GET with path param (auto-detected) + optional query param."""
    return {"item_id": item_id, "q": q}


@router.put("/mixed/<int:item_id>", response_model=ItemOut)
def update_item(item_id: int, q: str | None = Query(None), item: ItemIn = Body()):
    """PUT with path param + explicit query param + required body.

    NOTE: q must use Query() explicitly — PUT's default location is body.
    Without Query(), q would be treated as a body field, not a query param.
    """
    return {"id": item_id, "name": item.name, "price": item.price, "q": q}



try:
    from flask_jeroboam._version import __version__ as __version__
except ImportError:  # pragma: no cover
    __version__ = "0.0.dev0"

from flask_jeroboam.blueprint import Blueprint as Blueprint
from flask_jeroboam.jeroboam import Jeroboam as Jeroboam
from flask_jeroboam.models import InboundModel as InboundModel
from flask_jeroboam.models import OutboundModel as OutboundModel
from flask_jeroboam.view_arguments.functions import Body as Body
from flask_jeroboam.view_arguments.functions import Cookie as Cookie
from flask_jeroboam.view_arguments.functions import File as File
from flask_jeroboam.view_arguments.functions import Form as Form
from flask_jeroboam.view_arguments.functions import Header as Header
from flask_jeroboam.view_arguments.functions import Path as Path
from flask_jeroboam.view_arguments.functions import Query as Query
from flask_jeroboam.wrapper import current_app as current_app



import pytest
from flask.testing import FlaskClient

from tests.app_test.models.inbound import OptionalModelIn

response_missing = {
    "detail": [
        {
            "loc": ["query", "query"],
            "msg": "Field required",
            "type": "missing",
        }
    ]
}

response_not_valid_int = {
    "detail": [
        {
            "loc": ["query", "query"],
            "msg": "Input should be a valid integer, unable to parse string as an integer",
            "type": "int_parsing",
        }
    ]
}


def _valid(value) -> dict:
    """Valid function."""
    return {"query": value}


@pytest.mark.parametrize(
    "url,expected_status,expected_response",
    [
        ("/query", 400, response_missing),
        ("/query?query=baz", 200, _valid("baz")),
        ("/query?not_declared=baz", 400, response_missing),
        ("/query/optional", 200, _valid(None)),
        ("/query/optional?query=baz", 200, _valid("baz")),
        ("/query/optional?not_declared=baz", 200, _valid(None)),
        ("/query/int", 400, response_missing),
        ("/query/int?query=42", 200, _valid(42)),
        ("/query/int?query=42.5", 400, response_not_valid_int),
        ("/query/int?query=baz", 400, response_not_valid_int),
        ("/query/int?not_declared=baz", 400, response_missing),
        ("/query/int/optional", 200, _valid(None)),
        ("/query/int/optional?query=50", 200, _valid(50)),
        ("/query/int/optional?query=foo", 400, response_not_valid_int),
        ("/query/int/default", 200, _valid(10)),
        ("/query/int/default?query=50", 200, _valid(50)),
        ("/query/int/default?query=foo", 400, response_not_valid_int),
        ("/query/param", 200, _valid(None)),
        ("/query/param?query=50", 200, _valid("50")),
        ("/query/param-required", 400, response_missing),
        ("/query/param-required?query=50", 200, _valid("50")),
        ("/query/param-required/int", 400, response_missing),
        ("/query/param-required/int?query=50", 200, _valid(50)),
        ("/query/param-required/int?query=foo", 400, response_not_valid_int),
        ("/query/frozenset?query=1&query=1&query=2", 200, _valid("1,2")),
    ],
)
def test_get_query_operations(client, url, expected_status, expected_response):
    """Testing Various GET operations with query parameters.

    GIVEN a GET endpoint configiured with query parameters
    WHEN a request is made to the endpoint
    THEN the request is parsed and validated accordingly
    """
    response = client.get(url)
    assert response.json == expected_response
    assert response.status_code == expected_status


def test_valid_base_model_as_query_parameter(
    client: FlaskClient,
):
    """GIVEN a GET endpoint with a BaseModel as QueryParam
    WHEN hit with a valid query string
    THEN the parsed input is injected into the view function.
    """
    response = client.get("/query/base_model?page=1&type=item")
    assert response.status_code == 200
    assert response.json == {"page": 1, "type": "item"}


def test_valid_base_model_as_forwarded_query_parameter(
    client: FlaskClient,
):
    """GIVEN a GET endpoint with a BaseModel as Forward Ref QueryParam
    WHEN hit with a valid query string
    THEN the parsed input is injected into the view function.
    """
    response = client.get("/query/base_model/forward_ref?page=1&type=item")
    assert response.status_code == 200
    assert response.json == {"page": 1, "type": "item"}


def test_invalid_query_string_raise_400(
    client: FlaskClient,
):
    """GIVEN a GET endpoint with properly annotated argument
    WHEN hit with invalid queryString
    THEN the endpoint raise a 400 InvalidRequest Error
    """
    response = client.get("/query/base_model?page=not_a_valid_param")
    assert response.status_code == 400
    errors = response.json.get("detail", [])
    locs = [loc for e in errors for loc in e.get("loc", [])]
    assert "page" in locs
    assert "type" in locs


def test_query_string_for_list_arguments(
    client: FlaskClient,
):
    """GIVEN a GET endpoint with list arguments
    WHEN hit with proper formatted queryString
    THEN the arguments get injected into a Array
    """
    response = client.get("/query/list_of_strings?id[]=1&id[]=2")
    assert response.json == {"id": [1, 2]}
    assert response.status_code == 200


def test_query_optionnal_base_model(
    client: FlaskClient,
):
    """GIVEN an endpoint with Optionnal typed argument with Optional fields
    WHEN hit with an empty querystring
    THEN the endpoint is properly executed
    """
    response = client.get("/query/optional_model")
    assert response.status_code == 200
    assert response.json == OptionalModelIn().model_dump()



"""Jeroboam Test App factory."""

import os

from toml import load as load_toml  # type: ignore

from flask_jeroboam import Jeroboam
from tests.app_test.apps.body import router as body_router
from tests.app_test.apps.cookies import router as cookies_router
from tests.app_test.apps.file import router as file_router
from tests.app_test.apps.form import router as form_router
from tests.app_test.apps.header import router as header_router
from tests.app_test.apps.misc import router as misc_router
from tests.app_test.apps.mixed import router as mixed_router
from tests.app_test.apps.openapi import router as openapi_test_router
from tests.app_test.apps.outbound import router as outbound_router
from tests.app_test.apps.path import router as path_router
from tests.app_test.apps.query import router as query_router


def create_test_app():
    """Jeroboam test app factory."""
    app = Jeroboam("jeroboam_test", root_path=os.path.dirname(__file__))
    app.config.update(
        TESTING=True,
    )
    app.config.from_file("openapi.toml", load=load_toml)

    @app.route("/api_route", tags=["Base"])
    def non_operation():
        return {"message": "Hello World"}

    @app.get("/text", tags=["Base"])
    def get_text():
        return "Hello World"

    def non_decorated_route():
        return {"message": "Hello World"}

    app.add_url_rule("/non_decorated_route", view_func=non_decorated_route)

    app.register_blueprint(body_router)
    app.register_blueprint(cookies_router)
    app.register_blueprint(file_router)
    app.register_blueprint(form_router)
    app.register_blueprint(header_router)
    app.register_blueprint(path_router)
    app.register_blueprint(query_router)
    app.register_blueprint(misc_router)
    app.register_blueprint(mixed_router)
    app.register_blueprint(outbound_router)
    app.register_blueprint(openapi_test_router)

    app.init_app()

    return app



"""Testing Utils."""

from functools import partial
from typing import Optional, Union

import pytest
from flask.testing import FlaskClient
from pydantic import BaseModel, ValidationError
from pydantic_core import PydanticUndefined

from flask_jeroboam import Body
from flask_jeroboam._outboundhandler import OutboundHandler
from flask_jeroboam._utils import (
    _lenient_issubclass,
    _rename_query_params_keys,
    _unwrap_optional,
    get_typed_return_annotation,
    is_sequence_field,
)
from flask_jeroboam.datastructures import UploadFile
from flask_jeroboam.jeroboam import Jeroboam
from flask_jeroboam.view_arguments.arguments import (
    ArgumentLocation,
    BodyArgument,
    FormArgument,
    QueryArgument,
)
from flask_jeroboam.view_arguments.solved import SolvedArgument
from tests.app_test.models.inbound import ModelWithListIn
from tests.app_test.models.outbound import ModelWithListOut


def test_pascal_case_in_and_out_snake_case(
    one_shot_app: Jeroboam, one_shot_client: FlaskClient
):
    """GIVEN an endpoint with param typed with a Parser and response_model a Serializer
    WHEN payload is send in pascalCase
    THEN it lives in python in snake_case and send back in pascalCase
    """
    # We need to define the endpoint here to set the query_string_key_transformer first.
    one_shot_app.query_string_key_transformer = partial(
        _rename_query_params_keys, pattern=r"(.*)\[(.+)\]$"
    )

    @one_shot_app.get(
        "/query/special_pattern/after_configuration", response_model=ModelWithListOut
    )
    def read_items(payload: ModelWithListIn):
        return payload

    response = one_shot_client.get(
        "/query/special_pattern/after_configuration?page=1&perPage=10&id[]=1&id[]=2&order[name]=asc&order[age]=desc"
    )

    assert response.status_code == 200
    assert response.json == {
        "page": 1,
        "perPage": 10,
        "ids": [1, 2],
        "order": [{"name": "asc"}, {"age": "desc"}],
    }


def test_pascal_case_in_and_out_snake_case_without_transformer(
    one_shot_app: Jeroboam, one_shot_client: FlaskClient
):
    """GIVEN an endpoint with param typed with a Parser and response_model a Serializer
    WHEN payload is send in pascalCase
    THEN it lives in python in snake_case and send back in pascalCase
    """

    @one_shot_app.get(
        "/query/special_pattern/after_configuration", response_model=ModelWithListOut
    )
    def read_items(payload: ModelWithListIn):
        return payload

    response = one_shot_client.get(
        "/query/special_pattern?page=1&perPage=10&id[]=1&id[]=2&order[name]=asc&order[age]=desc"
    )

    assert response.status_code == 400
    errors = response.json.get("detail", [])
    assert len(errors) == 1
    # loc may use field name "order" (v2) or alias "order[]" (v1)
    assert any("order" in str(e.get("loc", "")) for e in errors)
    # msg may have "Value error, " prefix in v2
    assert any("Order must have at least 1 value" in e.get("msg", "") for e in errors)


def test_view_param_str_repr():
    """Test an internal function of ViewParameter."""
    param = Body("MyDefautValue")
    assert param.__repr__() == "BodyArgument(MyDefautValue)"


def test_solved_param_erroring():
    """Test an internal function of SolvedParameter."""
    param = Body("MyDefautValue")
    param.location = None
    solved_param = SolvedArgument(
        name="FaultySolvedParam", annotation=str, field_info=param
    )
    with pytest.raises(NotImplementedError):
        solved_param._get_values()


def test_file_upload():
    """Test UploadFile DataStructure raise an error if not a FileStorage."""

    class TestModel(BaseModel):
        file: UploadFile

    with pytest.raises(ValidationError):
        TestModel(file="BytesIO(b'Hello World !!')")  # type: ignore


def test_warn_on_two_identical_operation_id(one_shot_app: Jeroboam):
    """Test that a warning is raised when two identical operation_id are found."""
    app = Jeroboam(__name__)
    app.init_app()

    @app.get("/test")
    def test():
        pass

    @app.get("/test")
    def get_test():
        pass

    with pytest.warns(UserWarning):
        app.openapi  # type: ignore


# --- is_sequence_field ---


def test_is_sequence_field_with_no_annotation():
    """is_sequence_field returns False when the field has no annotation attribute."""
    from types import SimpleNamespace

    field = SimpleNamespace()  # no .annotation attribute
    assert is_sequence_field(field) is False


# --- ViewArgument.in_body ---


def test_in_body_true_for_body_argument():
    """BodyArgument.in_body is True."""
    assert BodyArgument().in_body is True


def test_in_body_true_for_form_argument():
    """FormArgument.in_body is True."""
    assert FormArgument().in_body is True


def test_in_body_false_for_query_argument():
    """QueryArgument.in_body is False."""
    assert QueryArgument().in_body is False


# --- SolvedArgument without field_info ---


def test_solved_argument_without_field_info():
    """SolvedArgument with field_info=None uses annotation directly (no Annotated wrapper)."""
    solved = SolvedArgument(name="x", annotation=str)
    assert solved.field_info is None
    # TypeAdapter should still work — no constraints to apply
    assert solved._type_adapter.validate_python("hello") == "hello"


# --- SolvedArgument.specialize with Ellipsis default ---


def test_specialize_normalises_ellipsis_default():
    """specialize() converts Ellipsis default to PydanticUndefined."""
    from types import SimpleNamespace

    view_param = SimpleNamespace(
        default=Ellipsis,
        location=ArgumentLocation.query,
        alias=None,
        embed=False,
        include_in_schema=True,
    )
    solved = SolvedArgument.specialize(name="x", annotation=str, view_param=view_param)
    assert solved.default is PydanticUndefined


# --- _unwrap_optional in _utils.py ---


def test_unwrap_optional_with_typing_optional():
    """Optional[int] (= Union[int, None]) unwraps to int via the typing.Union path."""
    assert _unwrap_optional(Optional[int]) is int  # noqa: UP045


def test_unwrap_optional_with_multi_union():
    """Union[int, str, None] has two non-None members; returned unchanged."""
    result = _unwrap_optional(Union[int, str, None])  # noqa: UP007
    assert result == Union[int, str, None]  # noqa: UP007


def test_unwrap_optional_non_union():
    """Plain int is returned unchanged."""
    assert _unwrap_optional(int) is int


# --- get_typed_return_annotation in _utils.py ---


def test_get_typed_return_annotation_returns_none_on_unresolvable_annotation():
    """Unresolvable forward reference causes get_type_hints to raise; returns None."""

    def func() -> "NonExistentType":  # type: ignore[name-defined]  # noqa: F821
        pass

    assert get_typed_return_annotation(func) is None


# --- OutboundHandler._adapt_datastructure_of with list ---


def test_adapt_datastructure_of_list():
    """A list of BaseModel instances is recursively adapted to a list of dicts."""

    class Item(BaseModel):
        name: str

    def view():
        pass

    handler = OutboundHandler(view, None, "GET", {})
    result = handler._adapt_datastructure_of([Item(name="a"), Item(name="b")])
    assert result == [{"name": "a"}, {"name": "b"}]


# --- _lenient_issubclass ---


def test_lenient_issubclass_returns_false_on_non_class():
    """_lenient_issubclass returns False when issubclass would raise TypeError."""
    assert _lenient_issubclass(42, int) is False



"""Path parameters - GET and POST endpoints."""

from flask_jeroboam import Jeroboam

app = Jeroboam("Jeroboam Inbound Features App")
app.init_app()


@app.get("/item/<int:id>/implicit")
def get_path_parameter(id: int):
    return f"Received id Argument is : {id}"


@app.post("/item/<int:id>/implicit")
def post_path_parameter(id: int):
    return f"Received id Argument is : {id}"


if __name__ == "__main__":
    app.run(host="localhost", port=5000)



"""Testing endpoints that combine multiple parameter locations.

These tests are critical for migration safety: the parameter routing logic
(path auto-detection, verb-based defaults, explicit Body()/Query()) must all
work together correctly. Regressions here mean parameters end up in the wrong
location or are silently ignored.

Key design behavior documented here:
- GET: all undecorated params default to QUERY location
- PUT/POST: all undecorated params default to BODY location
- Path params are auto-detected from the URL rule (no annotation needed)
- Explicit Query() / Body() always overrides the verb-based default
"""

import pytest
from flask.testing import FlaskClient

from flask_jeroboam import Jeroboam

# --- GET: path + optional query ---


@pytest.mark.parametrize(
    "url,expected_status,expected_body",
    [
        ("/mixed/42", 200, {"item_id": 42, "q": None}),
        ("/mixed/42?q=search", 200, {"item_id": 42, "q": "search"}),
        ("/mixed/99?q=hello", 200, {"item_id": 99, "q": "hello"}),
    ],
)
def test_get_with_path_and_optional_query(
    client: FlaskClient, url, expected_status, expected_body
):
    """GIVEN a GET endpoint with a path param and an optional query param
    WHEN called with various combinations
    THEN each parameter is resolved from the correct source.
    """
    response = client.get(url)
    assert response.status_code == expected_status
    assert response.json == expected_body


def test_get_path_param_wrong_type_returns_400(client: FlaskClient):
    """GIVEN a GET endpoint with an int path param
    WHEN called with a non-integer value
    THEN Flask's converter returns 404 before Jeroboam validation runs.
    """
    response = client.get("/mixed/not_an_int")
    assert response.status_code == 404


# --- PUT: path + optional query + required body ---


def test_put_with_all_three_locations(client: FlaskClient):
    """GIVEN a PUT endpoint with path, query and body params
    WHEN called with all three present
    THEN each is extracted from its correct source and the response model is applied.
    """
    response = client.put(
        "/mixed/7?q=filtered",
        json={"name": "Widget", "price": 9.99},
    )
    assert response.status_code == 201
    assert response.json == {"id": 7, "name": "Widget", "price": 9.99, "q": "filtered"}


def test_put_without_optional_query(client: FlaskClient):
    """GIVEN a PUT endpoint with an optional query param
    WHEN the query param is absent
    THEN it defaults to None without error.
    """
    response = client.put(
        "/mixed/3",
        json={"name": "Gadget", "price": 4.50},
    )
    assert response.status_code == 201
    assert response.json == {"id": 3, "name": "Gadget", "price": 4.50, "q": None}


def test_put_with_missing_required_body_field_returns_400(client: FlaskClient):
    """GIVEN a PUT endpoint with a required body model
    WHEN the body is missing a required field
    THEN a 400 validation error is returned.
    """
    response = client.put(
        "/mixed/1",
        json={"name": "Incomplete"},  # missing price
    )
    assert response.status_code == 400
    data = response.json
    assert data is not None
    assert "detail" in data
    errors = data["detail"]
    # Error loc: ['body', 'item', 'price'] — body param named 'item', field 'price'
    assert any("price" in e.get("loc", []) for e in errors)


def test_put_with_wrong_body_field_type_returns_400(client: FlaskClient):
    """GIVEN a PUT endpoint with a float body field
    WHEN the body contains a non-numeric string for that field
    THEN a 400 validation error is returned and the path param is not confused with the body.
    """
    response = client.put(
        "/mixed/5",
        json={"name": "Widget", "price": "not_a_float"},
    )
    assert response.status_code == 400
    assert response.json is not None
    assert "detail" in response.json


def test_put_body_fields_are_not_confused_with_query_params(client: FlaskClient):
    """GIVEN a PUT endpoint where q is declared with Query() and item with Body()
    WHEN q is in the query string and item fields are in the JSON body
    THEN each is extracted from the correct location.
    """
    response = client.put(
        "/mixed/10?q=from_query",
        json={"name": "from_body", "price": 1.0},
    )
    assert response.status_code == 201
    assert response.json["q"] == "from_query"
    assert response.json["name"] == "from_body"


def test_put_implicit_default_is_body_not_query(one_shot_app: Jeroboam):
    """GIVEN a PUT endpoint with an undecorated optional param (no Query() or Body())
    WHEN the param is passed in the query string
    THEN it is NOT found — PUT's implicit default is body, not query.

    This documents the verb-based default behavior: to use a query param on
    a PUT endpoint you must be explicit with Query().
    """

    @one_shot_app.put("/test_implicit_put/<int:item_id>")
    def implicit_put(item_id: int, q: str | None = None):
        # q defaults to body location on PUT — query string is ignored
        return {"item_id": item_id, "q": q}

    client = one_shot_app.test_client()
    # Sending q in the query string — it will NOT be parsed (body default)
    response = client.put("/test_implicit_put/1?q=ignored", json={})
    assert response.status_code == 201
    assert response.json["q"] is None  # q was not found in the body



"""Register a view function."""

from flask_jeroboam import Jeroboam

app = Jeroboam("Jeroboam Getting Started App")
app.init_app()


@app.get("/health")
def get_health():
    return {"status": "Ok"}


if __name__ == "__main__":
    app.run(host="localhost", port=5000)



"""Types used in the package."""

from collections.abc import Awaitable, Callable, Iterator, Mapping
from typing import (
    TYPE_CHECKING,
    Any,
    Protocol,
    TypeVar,
    Union,
)

from flask.typing import HeadersValue
from pydantic import BaseModel

if TYPE_CHECKING:  # pragma: no cover
    from _typeshed.wsgi import WSGIApplication  # noqa: F401
    from werkzeug.datastructures import Headers  # noqa: F401
    from werkzeug.wrappers import Response


class DataclassProtocol(Protocol):
    """A Protiocol to type annotate dataclasses."""

    __dataclass_fields__: dict
    __dataclass_params__: dict
    __post_init__: Callable | None


ResponseModel = type[BaseModel]
TypedParams = dict[str, Any]


DataClassType = TypeVar("DataClassType", bound=DataclassProtocol)


JeroboamBodyType = Union[
    "Response",
    DataClassType,
    BaseModel,
    str,
    bytes,
    list[Any],
    list[BaseModel],
    # Only dict is actually accepted, but Mapping allows for TypedDic
    Mapping[str, Any],
    Iterator[str],
    Iterator[bytes],
    DataclassProtocol,
]

JeroboamResponseWithStatusCode = (
    tuple[JeroboamBodyType, int] | tuple[JeroboamBodyType, int, HeadersValue]
)

JeroboamResponseReturnValue = Union[
    JeroboamBodyType,
    tuple[JeroboamBodyType, HeadersValue],
    tuple[JeroboamBodyType, int],
    tuple[JeroboamBodyType, int, HeadersValue],
    "WSGIApplication",
]

JeroboamReturnValue = (
    JeroboamResponseReturnValue | Awaitable[JeroboamResponseReturnValue]
)

JeroboamRouteCallable = (
    Callable[..., JeroboamResponseReturnValue]
    | Callable[..., Awaitable[JeroboamResponseReturnValue]]
)



"""Customized exceptions for the Flask-Jeroboam package.

They are small wrappers around werkzeug HTTP exceptions that customize
how the message is colllected and formatted.
"""

from typing import TYPE_CHECKING, Any

from werkzeug.exceptions import BadRequest, InternalServerError, NotFound

if TYPE_CHECKING:  # pragma: no cover
    from flask_jeroboam.jeroboam import Jeroboam


class RessourceNotFound(NotFound):
    """A slightly modified version of Werkzeug's RessourceNotFound Exception."""

    def __init__(
        self,
        msg: str | None = None,
        ressource_name: str | None = None,
        context: str | None = None,
    ):
        self.msg = msg
        self.ressource_name = ressource_name
        self.context = context
        self.response = None

    def __str__(self) -> str:
        return f"RessourceNotFound: {self.message}"

    @property
    def message(self) -> str:
        """Either format message or pass the msg property."""
        if self.msg is None:
            return f"{self.ressource_name} not found : {self.context}."
        return self.msg

    def handle(self) -> tuple[str, int]:
        """Handle the exception and return a message to the user."""
        return str(self), 404


class InvalidRequest(BadRequest):
    """A slightly modified version of Werkzeug's BadRequest Exception."""

    def __init__(self, errors: list[dict], *, body: Any = None):
        self.error_details = errors
        self.body = body
        self.response = None

    def handle(self) -> tuple[dict, int]:
        """Handle the exception and return a message to the user."""
        return {"detail": self.error_details}, 400


class ServerError(InternalServerError):
    """A slightly modifiedversion of Werkzeug's InternalServerError Exception."""

    def __init__(
        self, msg: str, error: Exception, trace: str, context: str | None = None
    ):
        self.msg = msg
        self.error = error
        self.trace = trace
        self.context = context
        self.response = None

    def __str__(self) -> str:
        return f"InternalServerError: {self.msg}"

    def handle(self) -> tuple[str, int]:
        """Handle the exception and return a message to the user."""
        return str(self), 500


class ResponseValidationError(ServerError):
    """When an error occurs on Outbound Validation."""

    def __str__(self) -> str:
        return f"InternalServerError: {self.msg}"


class JeroboamError(Exception):
    """Base Exception for Flask-Jeroboam."""


def handle_404(e):
    """Simple Hanlder for 404 errors."""
    return {"message": "Not Found"}, 404


def handle_500(e):
    """Simple Hanlder for 500 errors."""
    return {"message": "Internal Error"}, 500


def register_error_handlers(app: "Jeroboam"):
    """Register error handlers for the app."""
    app.register_error_handler(InvalidRequest, InvalidRequest.handle)
    app.register_error_handler(RessourceNotFound, RessourceNotFound.handle)
    app.register_error_handler(ServerError, ServerError.handle)
    app.register_error_handler(ResponseValidationError, ResponseValidationError.handle)
    app.register_error_handler(404, handle_404)
    app.register_error_handler(500, handle_500)



"""Misc Testing for Flask-Jeroboam."""

from flask.testing import FlaskClient


def test_delete_method(client: FlaskClient):
    """GIIVEN an endpoint with a different verb than GET or POST
    WHEN hit
    THEN it works like a regular endpoint
    """
    response = client.delete("/delete")
    assert response.status_code == 204
    assert response.data == b""



import pytest

from flask_jeroboam import Form
from flask_jeroboam.jeroboam import Jeroboam


def test_form_param_on_get_raise_warning():
    """A warning is raised when a Form parameter is used on a GET view.

    GIVEN a GET view with a Form parameter
    WHEN registering the view
    THEN a warning is raised
    """
    with pytest.warns(UserWarning):
        app = Jeroboam(__name__)

        def form_on_get(not_allowed: str = Form(...)):
            return "OK"

        app.get("/form_on_get")(form_on_get)



from flask.testing import FlaskClient

unspported_use_case = {
    "/non_decorated_route": {
        "get": {
            "responses": {
                "200": {
                    "description": "Successful Response",
                    "content": {"application/json": {"schema": {}}},
                }
            },
            "summary": "Non Decorated Route",
            "operationId": "non_decorated_route_non_decorated_route_get",
        }
    }
}

openapi_schema = {
    "openapi": "3.0.2",
    "info": {"title": "FastAPI", "version": "0.1.0"},
    "paths": {
        "/api_route": {
            "get": {
                "responses": {
                    "200": {
                        "description": "Successful Response",
                        "content": {"application/json": {"schema": {}}},
                    }
                },
                "summary": "Non Operation",
                "operationId": "get_non_operation",
                "tags": ["Base"],
            }
        },
        "/text": {
            "get": {
                "responses": {
                    "200": {
                        "description": "Successful Response",
                        "content": {"application/json": {"schema": {}}},
                    }
                },
                "summary": "Get Text",
                "operationId": "get_text",
                "tags": ["Base"],
            }
        },
        "/path/{item_id}": {
            "get": {
                "responses": {
                    "200": {
                        "description": "Successful Response",
                        "content": {"application/json": {"schema": {}}},
                    },
                    "400": {
                        "description": "Validation Error",
                        "content": {
                            "application/json": {
                                "schema": {
                                    "$ref": "#/components/schemas/HTTPValidationError"
                                }
                            }
                        },
                    },
                },
                "summary": "Get Id",
                "operationId": "path_params_router_get_id",
                "parameters": [
                    {
                        "required": True,
                        "schema": {"title": "Item Id"},
                        "name": "item_id",
                        "in": "path",
                    }
                ],
                "tags": ["Path"],
            }
        },
        "/path/str/{item_id}": {
            "get": {
                "responses": {
                    "200": {
                        "description": "Successful Response",
                        "content": {"application/json": {"schema": {}}},
                    },
                    "400": {
                        "description": "Validation Error",
                        "content": {
                            "application/json": {
                                "schema": {
                                    "$ref": "#/components/schemas/HTTPValidationError"
                                }
                            }
                        },
                    },
                },
                "summary": "Get Str Id",
                "operationId": "path_params_router_get_str_id",
                "parameters": [
                    {
                        "required": True,
                        "schema": {"title": "Item Id", "type": "string"},
                        "name": "item_id",
                        "in": "path",
                    }
                ],
                "tags": ["Path"],
            }
        },
        "/path/int/{item_id}": {
            "get": {
                "responses": {
                    "200": {
                        "description": "Successful Response",
                        "content": {"application/json": {"schema": {}}},
                    },
                    "400": {
                        "description": "Validation Error",
                        "content": {
                            "application/json": {
                                "schema": {
                                    "$ref": "#/components/schemas/HTTPValidationError"
                                }
                            }
                        },
                    },
                },
                "summary": "Get Int Id",
                "operationId": "path_params_router_get_int_id",
                "parameters": [
                    {
                        "required": True,
                        "schema": {"title": "Item Id", "type": "integer"},
                        "name": "item_id",
                        "in": "path",
                    }
                ],
                "tags": ["Path"],
            }
        },
        "/path/float/{item_id}": {
            "get": {
                "responses": {
                    "200": {
                        "description": "Successful Response",
                        "content": {"application/json": {"schema": {}}},
                    },
                    "400": {
                        "description": "Validation Error",
                        "content": {
                            "application/json": {
                                "schema": {
                                    "$ref": "#/components/schemas/HTTPValidationError"
                                }
                            }
                        },
                    },
                },
                "summary": "Get Float Id",
                "operationId": "path_params_router_get_float_id",
                "parameters": [
                    {
                        "required": True,
                        "schema": {"title": "Item Id", "type": "number"},
                        "name": "item_id",
                        "in": "path",
                    }
                ],
                "tags": ["Path"],
            }
        },
        "/path/bool/{item_id}": {
            "get": {
                "responses": {
                    "200": {
                        "description": "Successful Response",
                        "content": {"application/json": {"schema": {}}},
                    },
                    "400": {
                        "description": "Validation Error",
                        "content": {
                            "application/json": {
                                "schema": {
                                    "$ref": "#/components/schemas/HTTPValidationError"
                                }
                            }
                        },
                    },
                },
                "summary": "Get Bool Id",
                "operationId": "path_params_router_get_bool_id",
                "parameters": [
                    {
                        "required": True,
                        "schema": {"title": "Item Id", "type": "boolean"},
                        "name": "item_id",
                        "in": "path",
                    }
                ],
                "tags": ["Path"],
            }
        },
        "/path/param/{item_id}": {
            "get": {
                "responses": {
                    "200": {
                        "description": "Successful Response",
                        "content": {"application/json": {"schema": {}}},
                    },
                    "400": {
                        "description": "Validation Error",
                        "content": {
                            "application/json": {
                                "schema": {
                                    "$ref": "#/components/schemas/HTTPValidationError"
                                }
                            }
                        },
                    },
                },
                "summary": "Get Path Param Id",
                "operationId": "path_params_router_get_path_param_id",
                "parameters": [
                    {
                        "required": True,
                        "schema": {"title": "Item Id", "type": "string"},
                        "name": "item_id",
                        "in": "path",
                    }
                ],
                "tags": ["Path"],
            }
        },
        "/path/param-required/{item_id}": {
            "get": {
                "responses": {
                    "200": {
                        "description": "Successful Response",
                        "content": {"application/json": {"schema": {}}},
                    },
                    "400": {
                        "description": "Validation Error",
                        "content": {
                            "application/json": {
                                "schema": {
                                    "$ref": "#/components/schemas/HTTPValidationError"
                                }
                            }
                        },
                    },
                },
                "summary": "Get Path Param Required Id",
                "operationId": "path_params_router_get_path_param_required_id",
                "parameters": [
                    {
                        "required": True,
                        "schema": {"title": "Item Id", "type": "string"},
                        "name": "item_id",
                        "in": "path",
                    }
                ],
                "tags": ["Path"],
            }
        },
        "/path/param-minlength/{item_id}": {
            "get": {
                "responses": {
                    "200": {
                        "description": "Successful Response",
                        "content": {"application/json": {"schema": {}}},
                    },
                    "400": {
                        "description": "Validation Error",
                        "content": {
                            "application/json": {
                                "schema": {
                                    "$ref": "#/components/schemas/HTTPValidationError"
                                }
                            }
                        },
                    },
                },
                "summary": "Get Path Param Min Length",
                "operationId": "path_params_router_get_path_param_min_length",
                "parameters": [
                    {
                        "required": True,
                        "schema": {
                            "title": "Item Id",
                            "minLength": 3,
                            "type": "string",
                        },
                        "name": "item_id",
                        "in": "path",
                    }
                ],
                "tags": ["Path"],
            }
        },
        "/path/param-maxlength/{item_id}": {
            "get": {
                "responses": {
                    "200": {
                        "description": "Successful Response",
                        "content": {"application/json": {"schema": {}}},
                    },
                    "400": {
                        "description": "Validation Error",
                        "content": {
                            "application/json": {
                                "schema": {
                                    "$ref": "#/components/schemas/HTTPValidationError"
                                }
                            }
                        },
                    },
                },
                "summary": "Get Path Param Max Length",
                "operationId": "path_params_router_get_path_param_max_length",
                "parameters": [
                    {
                        "required": True,
                        "schema": {
                            "title": "Item Id",
                            "maxLength": 3,
                            "type": "string",
                        },
                        "name": "item_id",
                        "in": "path",
                    }
                ],
                "tags": ["Path"],
            }
        },
        "/path/param-min_maxlength/{item_id}": {
            "get": {
                "responses": {
                    "200": {
                        "description": "Successful Response",
                        "content": {"application/json": {"schema": {}}},
                    },
                    "400": {
                        "description": "Validation Error",
                        "content": {
                            "application/json": {
                                "schema": {
                                    "$ref": "#/components/schemas/HTTPValidationError"
                                }
                            }
                        },
                    },
                },
                "summary": "Get Path Param Min Max Length",
                "operationId": "path_params_router_get_path_param_min_max_length",
                "parameters": [
                    {
                        "required": True,
                        "schema": {
                            "title": "Item Id",
                            "maxLength": 3,
                            "minLength": 2,
                            "type": "string",
                        },
                        "name": "item_id",
                        "in": "path",
                    }
                ],
                "tags": ["Path"],
            }
        },
        "/path/param-gt/{item_id}": {
            "get": {
                "responses": {
                    "200": {
                        "description": "Successful Response",
                        "content": {"application/json": {"schema": {}}},
                    },
                    "400": {
                        "description": "Validation Error",
                        "content": {
                            "application/json": {
                                "schema": {
                                    "$ref": "#/components/schemas/HTTPValidationError"
                                }
                            }
                        },
                    },
                },
                "summary": "Get Path Param Gt",
                "operationId": "path_params_router_get_path_param_gt",
                "parameters": [
                    {
                        "required": True,
                        "schema": {
                            "title": "Item Id",
                            "exclusiveMinimum": 3.0,
                            "type": "number",
                        },
                        "name": "item_id",
                        "in": "path",
                    }
                ],
                "tags": ["Path"],
            }
        },
        "/path/param-gt0/{item_id}": {
            "get": {
                "responses": {
                    "200": {
                        "description": "Successful Response",
                        "content": {"application/json": {"schema": {}}},
                    },
                    "400": {
                        "description": "Validation Error",
                        "content": {
                            "application/json": {
                                "schema": {
                                    "$ref": "#/components/schemas/HTTPValidationError"
                                }
                            }
                        },
                    },
                },
                "summary": "Get Path Param Gt0",
                "operationId": "path_params_router_get_path_param_gt0",
                "parameters": [
                    {
                        "required": True,
                        "schema": {
                            "title": "Item Id",
                            "exclusiveMinimum": 0.0,
                            "type": "number",
                        },
                        "name": "item_id",
                        "in": "path",
                    }
                ],
                "tags": ["Path"],
            }
        },
        "/path/param-ge/{item_id}": {
            "get": {
                "responses": {
                    "200": {
                        "description": "Successful Response",
                        "content": {"application/json": {"schema": {}}},
                    },
                    "400": {
                        "description": "Validation Error",
                        "content": {
                            "application/json": {
                                "schema": {
                                    "$ref": "#/components/schemas/HTTPValidationError"
                                }
                            }
                        },
                    },
                },
                "summary": "Get Path Param Ge",
                "operationId": "path_params_router_get_path_param_ge",
                "parameters": [
                    {
                        "required": True,
                        "schema": {
                            "title": "Item Id",
                            "minimum": 3.0,
                            "type": "number",
                        },
                        "name": "item_id",
                        "in": "path",
                    }
                ],
                "tags": ["Path"],
            }
        },
        "/path/param-lt/{item_id}": {
            "get": {
                "responses": {
                    "200": {
                        "description": "Successful Response",
                        "content": {"application/json": {"schema": {}}},
                    },
                    "400": {
                        "description": "Validation Error",
                        "content": {
                            "application/json": {
                                "schema": {
                                    "$ref": "#/components/schemas/HTTPValidationError"
                                }
                            }
                        },
                    },
                },
                "summary": "Get Path Param Lt",
                "operationId": "path_params_router_get_path_param_lt",
                "parameters": [
                    {
                        "required": True,
                        "schema": {
                            "title": "Item Id",
                            "exclusiveMaximum": 3.0,
                            "type": "number",
                        },
                        "name": "item_id",
                        "in": "path",
                    }
                ],
                "tags": ["Path"],
            }
        },
        "/path/param-lt0/{item_id}": {
            "get": {
                "responses": {
                    "200": {
                        "description": "Successful Response",
                        "content": {"application/json": {"schema": {}}},
                    },
                    "400": {
                        "description": "Validation Error",
                        "content": {
                            "application/json": {
                                "schema": {
                                    "$ref": "#/components/schemas/HTTPValidationError"
                                }
                            }
                        },
                    },
                },
                "summary": "Get Path Param Lt0",
                "operationId": "path_params_router_get_path_param_lt0",
                "parameters": [
                    {
                        "required": True,
                        "schema": {
                            "title": "Item Id",
                            "exclusiveMaximum": 0.0,
                            "type": "number",
                        },
                        "name": "item_id",
                        "in": "path",
                    }
                ],
                "tags": ["Path"],
            }
        },
        "/path/param-le/{item_id}": {
            "get": {
                "responses": {
                    "200": {
                        "description": "Successful Response",
                        "content": {"application/json": {"schema": {}}},
                    },
                    "400": {
                        "description": "Validation Error",
                        "content": {
                            "application/json": {
                                "schema": {
                                    "$ref": "#/components/schemas/HTTPValidationError"
                                }
                            }
                        },
                    },
                },
                "summary": "Get Path Param Le",
                "operationId": "path_params_router_get_path_param_le",
                "parameters": [
                    {
                        "required": True,
                        "schema": {
                            "title": "Item Id",
                            "maximum": 3.0,
                            "type": "number",
                        },
                        "name": "item_id",
                        "in": "path",
                    }
                ],
                "tags": ["Path"],
            }
        },
        "/path/param-lt-gt/{item_id}": {
            "get": {
                "responses": {
                    "200": {
                        "description": "Successful Response",
                        "content": {"application/json": {"schema": {}}},
                    },
                    "400": {
                        "description": "Validation Error",
                        "content": {
                            "application/json": {
                                "schema": {
                                    "$ref": "#/components/schemas/HTTPValidationError"
                                }
                            }
                        },
                    },
                },
                "summary": "Get Path Param Lt Gt",
                "operationId": "path_params_router_get_path_param_lt_gt",
                "parameters": [
                    {
                        "required": True,
                        "schema": {
                            "title": "Item Id",
                            "exclusiveMaximum": 3.0,
                            "exclusiveMinimum": 1.0,
                            "type": "number",
                        },
                        "name": "item_id",
                        "in": "path",
                    }
                ],
                "tags": ["Path"],
            }
        },
        "/path/param-le-ge/{item_id}": {
            "get": {
                "responses": {
                    "200": {
                        "description": "Successful Response",
                        "content": {"application/json": {"schema": {}}},
                    },
                    "400": {
                        "description": "Validation Error",
                        "content": {
                            "application/json": {
                                "schema": {
                                    "$ref": "#/components/schemas/HTTPValidationError"
                                }
                            }
                        },
                    },
                },
                "summary": "Get Path Param Le Ge",
                "operationId": "path_params_router_get_path_param_le_ge",
                "parameters": [
                    {
                        "required": True,
                        "schema": {
                            "title": "Item Id",
                            "maximum": 3.0,
                            "minimum": 1.0,
                            "type": "number",
                        },
                        "name": "item_id",
                        "in": "path",
                    }
                ],
                "tags": ["Path"],
            }
        },
        "/path/param-lt-int/{item_id}": {
            "get": {
                "responses": {
                    "200": {
                        "description": "Successful Response",
                        "content": {"application/json": {"schema": {}}},
                    },
                    "400": {
                        "description": "Validation Error",
                        "content": {
                            "application/json": {
                                "schema": {
                                    "$ref": "#/components/schemas/HTTPValidationError"
                                }
                            }
                        },
                    },
                },
                "summary": "Get Path Param Lt Int",
                "operationId": "path_params_router_get_path_param_lt_int",
                "parameters": [
                    {
                        "required": True,
                        "schema": {
                            "title": "Item Id",
                            "exclusiveMaximum": 3.0,
                            "type": "integer",
                        },
                        "name": "item_id",
                        "in": "path",
                    }
                ],
                "tags": ["Path"],
            }
        },
        "/path/param-gt-int/{item_id}": {
            "get": {
                "responses": {
                    "200": {
                        "description": "Successful Response",
                        "content": {"application/json": {"schema": {}}},
                    },
                    "400": {
                        "description": "Validation Error",
                        "content": {
                            "application/json": {
                                "schema": {
                                    "$ref": "#/components/schemas/HTTPValidationError"
                                }
                            }
                        },
                    },
                },
                "summary": "Get Path Param Gt Int",
                "operationId": "path_params_router_get_path_param_gt_int",
                "parameters": [
                    {
                        "required": True,
                        "schema": {
                            "title": "Item Id",
                            "exclusiveMinimum": 3.0,
                            "type": "integer",
                        },
                        "name": "item_id",
                        "in": "path",
                    }
                ],
                "tags": ["Path"],
            }
        },
        "/path/param-le-int/{item_id}": {
            "get": {
                "responses": {
                    "200": {
                        "description": "Successful Response",
                        "content": {"application/json": {"schema": {}}},
                    },
                    "400": {
                        "description": "Validation Error",
                        "content": {
                            "application/json": {
                                "schema": {
                                    "$ref": "#/components/schemas/HTTPValidationError"
                                }
                            }
                        },
                    },
                },
                "summary": "Get Path Param Le Int",
                "operationId": "path_params_router_get_path_param_le_int",
                "parameters": [
                    {
                        "required": True,
                        "schema": {
                            "title": "Item Id",
                            "maximum": 3.0,
                            "type": "integer",
                        },
                        "name": "item_id",
                        "in": "path",
                    }
                ],
                "tags": ["Path"],
            }
        },
        "/path/param-ge-int/{item_id}": {
            "get": {
                "responses": {
                    "200": {
                        "description": "Successful Response",
                        "content": {"application/json": {"schema": {}}},
                    },
                    "400": {
                        "description": "Validation Error",
                        "content": {
                            "application/json": {
                                "schema": {
                                    "$ref": "#/components/schemas/HTTPValidationError"
                                }
                            }
                        },
                    },
                },
                "summary": "Get Path Param Ge Int",
                "operationId": "path_params_router_get_path_param_ge_int",
                "parameters": [
                    {
                        "required": True,
                        "schema": {
                            "title": "Item Id",
                            "minimum": 3.0,
                            "type": "integer",
                        },
                        "name": "item_id",
                        "in": "path",
                    }
                ],
                "tags": ["Path"],
            }
        },
        "/path/param-lt-gt-int/{item_id}": {
            "get": {
                "responses": {
                    "200": {
                        "description": "Successful Response",
                        "content": {"application/json": {"schema": {}}},
                    },
                    "400": {
                        "description": "Validation Error",
                        "content": {
                            "application/json": {
                                "schema": {
                                    "$ref": "#/components/schemas/HTTPValidationError"
                                }
                            }
                        },
                    },
                },
                "summary": "Get Path Param Lt Gt Int",
                "operationId": "path_params_router_get_path_param_lt_gt_int",
                "parameters": [
                    {
                        "required": True,
                        "schema": {
                            "title": "Item Id",
                            "exclusiveMaximum": 3.0,
                            "exclusiveMinimum": 1.0,
                            "type": "integer",
                        },
                        "name": "item_id",
                        "in": "path",
                    }
                ],
                "tags": ["Path"],
            }
        },
        "/path/param-le-ge-int/{item_id}": {
            "get": {
                "responses": {
                    "200": {
                        "description": "Successful Response",
                        "content": {"application/json": {"schema": {}}},
                    },
                    "400": {
                        "description": "Validation Error",
                        "content": {
                            "application/json": {
                                "schema": {
                                    "$ref": "#/components/schemas/HTTPValidationError"
                                }
                            }
                        },
                    },
                },
                "summary": "Get Path Param Le Ge Int",
                "operationId": "path_params_router_get_path_param_le_ge_int",
                "parameters": [
                    {
                        "required": True,
                        "schema": {
                            "title": "Item Id",
                            "maximum": 3.0,
                            "minimum": 1.0,
                            "type": "integer",
                        },
                        "name": "item_id",
                        "in": "path",
                    }
                ],
                "tags": ["Path"],
            }
        },
        "/query": {
            "get": {
                "responses": {
                    "200": {
                        "description": "Successful Response",
                        "content": {"application/json": {"schema": {}}},
                    },
                    "400": {
                        "description": "Validation Error",
                        "content": {
                            "application/json": {
                                "schema": {
                                    "$ref": "#/components/schemas/HTTPValidationError"
                                }
                            }
                        },
                    },
                },
                "summary": "Get Query",
                "operationId": "query_params_router_get_query",
                "parameters": [
                    {
                        "required": True,
                        "schema": {"title": "Query"},
                        "name": "query",
                        "in": "query",
                    }
                ],
                "tags": ["Query"],
            }
        },
        "/query/optional": {
            "get": {
                "responses": {
                    "200": {
                        "description": "Successful Response",
                        "content": {"application/json": {"schema": {}}},
                    },
                    "400": {
                        "description": "Validation Error",
                        "content": {
                            "application/json": {
                                "schema": {
                                    "$ref": "#/components/schemas/HTTPValidationError"
                                }
                            }
                        },
                    },
                },
                "summary": "Get Query Optional",
                "operationId": "query_params_router_get_query_optional",
                "parameters": [
                    {
                        "required": False,
                        "schema": {"title": "Query"},
                        "name": "query",
                        "in": "query",
                    }
                ],
                "tags": ["Query"],
            }
        },
        "/query/int": {
            "get": {
                "responses": {
                    "200": {
                        "description": "Successful Response",
                        "content": {"application/json": {"schema": {}}},
                    },
                    "400": {
                        "description": "Validation Error",
                        "content": {
                            "application/json": {
                                "schema": {
                                    "$ref": "#/components/schemas/HTTPValidationError"
                                }
                            }
                        },
                    },
                },
                "summary": "Get Query Type",
                "operationId": "query_params_router_get_query_type",
                "parameters": [
                    {
                        "required": True,
                        "schema": {"title": "Query", "type": "integer"},
                        "name": "query",
                        "in": "query",
                    }
                ],
                "tags": ["Query"],
            }
        },
        "/query/int/optional": {
            "get": {
                "responses": {
                    "200": {
                        "description": "Successful Response",
                        "content": {"application/json": {"schema": {}}},
                    },
                    "400": {
                        "description": "Validation Error",
                        "content": {
                            "application/json": {
                                "schema": {
                                    "$ref": "#/components/schemas/HTTPValidationError"
                                }
                            }
                        },
                    },
                },
                "summary": "Get Query Type Optional",
                "operationId": "query_params_router_get_query_type_optional",
                "parameters": [
                    {
                        "required": False,
                        "schema": {"title": "Query", "type": "integer"},
                        "name": "query",
                        "in": "query",
                    }
                ],
                "tags": ["Query"],
            }
        },
        "/query/int/default": {
            "get": {
                "responses": {
                    "200": {
                        "description": "Successful Response",
                        "content": {"application/json": {"schema": {}}},
                    },
                    "400": {
                        "description": "Validation Error",
                        "content": {
                            "application/json": {
                                "schema": {
                                    "$ref": "#/components/schemas/HTTPValidationError"
                                }
                            }
                        },
                    },
                },
                "summary": "Get Query Type Int Default",
                "operationId": "query_params_router_get_query_type_int_default",
                "parameters": [
                    {
                        "required": False,
                        "schema": {"title": "Query", "type": "integer", "default": 10},
                        "name": "query",
                        "in": "query",
                    }
                ],
                "tags": ["Query"],
            }
        },
        "/query/param": {
            "get": {
                "responses": {
                    "200": {
                        "description": "Successful Response",
                        "content": {"application/json": {"schema": {}}},
                    },
                    "400": {
                        "description": "Validation Error",
                        "content": {
                            "application/json": {
                                "schema": {
                                    "$ref": "#/components/schemas/HTTPValidationError"
                                }
                            }
                        },
                    },
                },
                "summary": "Get Query Param",
                "operationId": "query_params_router_get_query_param",
                "parameters": [
                    {
                        "required": False,
                        "schema": {"title": "Query"},
                        "name": "query",
                        "in": "query",
                    }
                ],
                "tags": ["Query"],
            }
        },
        "/query/param-required": {
            "get": {
                "responses": {
                    "200": {
                        "description": "Successful Response",
                        "content": {"application/json": {"schema": {}}},
                    },
                    "400": {
                        "description": "Validation Error",
                        "content": {
                            "application/json": {
                                "schema": {
                                    "$ref": "#/components/schemas/HTTPValidationError"
                                }
                            }
                        },
                    },
                },
                "summary": "Get Query Param Required",
                "operationId": "query_params_router_get_query_param_required",
                "parameters": [
                    {
                        "required": True,
                        "schema": {"title": "Query"},
                        "name": "query",
                        "in": "query",
                    }
                ],
                "tags": ["Query"],
            }
        },
        "/query/param-required/int": {
            "get": {
                "responses": {
                    "200": {
                        "description": "Successful Response",
                        "content": {"application/json": {"schema": {}}},
                    },
                    "400": {
                        "description": "Validation Error",
                        "content": {
                            "application/json": {
                                "schema": {
                                    "$ref": "#/components/schemas/HTTPValidationError"
                                }
                            }
                        },
                    },
                },
                "summary": "Get Query Param Required Type",
                "operationId": "query_params_router_get_query_param_required_type",
                "parameters": [
                    {
                        "required": True,
                        "schema": {"title": "Query", "type": "integer"},
                        "name": "query",
                        "in": "query",
                    }
                ],
                "tags": ["Query"],
            }
        },
        "/enum-status-code": {
            "get": {
                "responses": {
                    "201": {
                        "description": "Successful Response",
                        "content": {"application/json": {"schema": {}}},
                    },
                },
                "summary": "Get Enum Status Code",
                "operationId": "query_params_router_get_enum_status_code",
                "tags": ["Query"],
            }
        },
        "/query/frozenset": {
            "get": {
                "summary": "Get Query Type Frozenset",
                "operationId": "query_params_router_get_query_type_frozenset",
                "parameters": [
                    {
                        "required": True,
                        "schema": {
                            "title": "Query",
                            "uniqueItems": True,
                            "type": "array",
                            "items": {"type": "integer"},
                        },
                        "name": "query",
                        "in": "query",
                    }
                ],
                "responses": {
                    "200": {
                        "description": "Successful Response",
                        "content": {"application/json": {"schema": {}}},
                    },
                    "400": {
                        "description": "Validation Error",
                        "content": {
                            "application/json": {
                                "schema": {
                                    "$ref": "#/components/schemas/HTTPValidationError"
                                }
                            }
                        },
                    },
                },
                "tags": ["Query"],
            }
        },
    },
    "components": {
        "schemas": {
            "ValidationError": {
                "title": "ValidationError",
                "required": ["loc", "msg", "type"],
                "type": "object",
                "properties": {
                    "loc": {
                        "title": "Location",
                        "type": "array",
                        "items": {"anyOf": [{"type": "string"}, {"type": "integer"}]},
                    },
                    "msg": {"title": "Message", "type": "string"},
                    "type": {"title": "Error Type", "type": "string"},
                },
            },
            "HTTPValidationError": {
                "title": "HTTPValidationError",
                "type": "object",
                "properties": {
                    "detail": {
                        "title": "Detail",
                        "type": "array",
                        "items": {"$ref": "#/components/schemas/ValidationError"},
                    }
                },
            },
        }
    },
}


def test_paths_item(client: FlaskClient):
    """Test info endpoint."""
    response = client.get("/openapi.json")
    assert response.status_code == 200
    return_shema = response.json
    assert return_shema
    for path in openapi_schema["paths"]:  # type: ignore
        assert (
            return_shema["paths"][path] == openapi_schema["paths"][path]  # type: ignore
        ), f"{path} don't match"


def test_error_schemas(client: FlaskClient):
    """Test info endpoint."""
    response = client.get("/openapi.json")
    assert response.status_code == 200
    return_schema = response.json
    target_schema = openapi_schema["components"]["schemas"]  # type: ignore
    assert return_schema
    assert (
        return_schema["components"]["schemas"]["HTTPValidationError"]  # type: ignore
        == target_schema["HTTPValidationError"]  # type: ignore
    )
    assert (
        return_schema["components"]["schemas"]["ValidationError"]  # type: ignore
        == target_schema["ValidationError"]  # type: ignore
    )



"""The Scaffold Override.

It is used to override the route method of both Flask and Blueprints Object
in Jeroboam and jeroboam's Blueprints.
"""

from collections.abc import Callable
from typing import Any

from flask.sansio.scaffold import setupmethod
from typing_extensions import TypeVar

from flask_jeroboam.typing import JeroboamRouteCallable
from flask_jeroboam.view import JeroboamView

R = TypeVar("R", bound=Any)


class JeroboamScaffoldOverRide:
    """Mixin to override the route method.

    The route method is overriden by a custom flask_jeroboam
    route decorator.
    """

    @setupmethod
    def route(
        self, rule: str, **options: Any
    ) -> Callable[[JeroboamRouteCallable], JeroboamRouteCallable]:
        """View function decorator to register a route.

        Decorate a view function to register it with the given URL
        rule and options. Calls :meth:`add_url_rule`, which has more
        details about the implementation.

        .. code-block:: python

            @app.route("/")
            def index():
                return "Hello, World!"

        See :ref:`url-route-registrations`.

        The endpoint name for the route defaults to the name of the view
        function if the ``endpoint`` parameter isn't passed.

        The ``methods`` parameter defaults to ``["GET"]``. ``HEAD`` and
        ``OPTIONS`` are added automatically.

        :param rule: The URL rule string.
        :param options: Extra options passed to the
            :class:`~werkzeug.routing.Rule` object.
        """

        def decorator(view_func: JeroboamRouteCallable) -> JeroboamRouteCallable:
            options.setdefault("tags", []).extend(getattr(self, "tags", []))
            blueprint_option = getattr(self, "include_in_openapi", None)
            view = JeroboamView(rule, view_func, options) if view_func else None
            options["main_method"] = getattr(view, "main_http_verb", None)
            if blueprint_option is not None:
                options["include_in_openapi"] = blueprint_option
            self.add_url_rule(  # type: ignore
                rule,
                view.endpoint,
                view.as_view,
                **options,  # type: ignore
            )
            return view_func

        return decorator



"""Main builder function for OPENAPI schema."""

from typing import TYPE_CHECKING, Any

from flask_jeroboam._utils import _memoized_update_if_value
from flask_jeroboam.openapi._utils import (
    _build_openapi_path_item,
    _get_flat_models_from_jeroboam_views,
    _get_model_definitions,
)
from flask_jeroboam.openapi.models.openapi import Components, Info, OpenAPI, Tag

if TYPE_CHECKING:  # pragma: no cover
    from flask_jeroboam.jeroboam import Jeroboam
    from flask_jeroboam.rule import JeroboamRule
    from flask_jeroboam.view import JeroboamView


def _collect_jeroboam_views(
    app: "Jeroboam", rules: list["JeroboamRule"]
) -> "list[JeroboamView | None]":
    """Return the JeroboamView attached to each rule's endpoint, or None."""
    return [
        getattr(app.view_functions[rule.endpoint], "__jeroboam_view__", None)
        for rule in rules
    ]


def _filter_definitions(definitions: dict[str, Any]) -> dict[str, Any]:
    """Sort and strip synthetic request-body wrapper models from definitions."""
    return {
        k: definitions[k]
        for k in sorted(definitions)
        if not k.endswith("request_body_as_model")
    }


def build_openapi(
    *,
    app: "Jeroboam",
    rules: list["JeroboamRule"],
    tags: list[dict[str, Any]] | None = None,
) -> OpenAPI:
    """Generate an OpenAPI schema for the given routes.

    TODO: Gérer les securitySchemes.
    Credits: Refactoring of FastApi's get_openapi.
    """
    # Meta
    openapi_version = app.config.get("JEROBOAM_OPENAPI_VERSION", "3.0.2")
    info = Info.model_validate(
        {
            "title": app.config.get("JEROBOAM_TITLE", None) or app.name,
            "version": app.config.get("JEROBOAM_VERSION", "0.1.0"),
            "description": app.config.get("JEROBOAM_DESCRIPTION", None),
            "termsOfService": app.config.get("JEROBOAM_TERMS_OF_SERVICE", None),
            "contact": app.config.get("JEROBOAM_CONTACT", None),
            "license": app.config.get("JEROBOAM_LICENCE_INFO", None),
        }
    )
    servers = app.config.get("JEROBOAM_SERVERS", None)

    # Préparation
    paths: dict[str, Any] = {}
    components: dict[str, dict[str, Any]] = {}
    operation_ids: set[str] = set()

    # Les jerobomas views, probablement à déléguer à l'app
    jeroboam_views = _collect_jeroboam_views(app, rules)

    # On créer des objects intermédiaires
    flat_models = _get_flat_models_from_jeroboam_views(jeroboam_views, rules)
    definitions = _get_model_definitions(
        flat_models=flat_models,
    )

    # On itères sur les rules and views pour récuperr les Paths Items et Définitions.
    for rule, jeroboam_view in zip(rules, jeroboam_views):
        if rule.include_in_openapi is False or jeroboam_view is None:
            continue
        path_dict, security_schemes, path_definitions = _build_openapi_path_item(
            rule=rule,
            jeroboam_view=jeroboam_view,
            operation_ids=operation_ids,
        )
        _memoized_update_if_value(rule.openapi_path, path_dict, paths)
        # TODO: ici il faut être capable de récupérer les securitySchemes
        _memoized_update_if_value("security_schemes", security_schemes, components)
        definitions.update(path_definitions or {})

    definitions = _filter_definitions(definitions)
    # On package le tout.
    return OpenAPI(
        openapi=openapi_version,
        info=info,
        servers=servers,
        paths=paths,
        tags=[Tag(**tag) for tag in tags or []],
        components=Components(schemas=definitions),
    )



"""Implicit location - GET and POST endpoints."""

from flask_jeroboam import Jeroboam

app = Jeroboam("Jeroboam Inbound Features App")
app.init_app()


@app.get("/implicit_location_is_query_string")
def get_implicit_argument(page: int):
    return f"Received Page Argument is : {page}"


@app.post("/implicit_location_is_body")
def post_implicit_argument(page: int):
    return f"Received Page Argument is : {page}"


if __name__ == "__main__":
    app.run(host="localhost", port=5000)



"""Code Sample for Outbound Features Documentation."""

from pydantic import BaseModel, Field

from flask_jeroboam import Jeroboam

app = Jeroboam("Jeroboam Outbound Features App")
app.init_app()


class Task(BaseModel):
    id: int
    name: str
    description: str | None = Field("Just here to make a point.")


class TaskIn(BaseModel):
    name: str
    description: str | None = None


class TaskOut(TaskIn):
    id: int


@app.get("/tasks/<int:task_id>", response_model=Task)
def read_explicit_task(task_id: int):
    return {"id": task_id, "name": "Find the answer."}


@app.get("/tasks/<int:task_id>/no_response_model")
def read_task_dictionnary_without_response_mode(task_id: int):
    return {"id": task_id, "name": "I'm from the dictionary."}


@app.get("/tasks/<int:task_id>/implicit_from_annotation")
def read_implicit_task(task_id: int) -> Task:
    return Task.model_validate({"id": task_id, "name": "Implicit from Annotation"})


@app.get("/tasks/<int:task_id>/implicit_no_annotation")
def read_implicit_no_annotation(task_id: int):
    return Task.model_validate({"id": task_id, "name": "Implicit from Annotation"})


@app.get("/tasks/<int:task_id>/response_model_off", response_model=None)
def read_response_model_off(task_id: int):
    return {"id": task_id, "name": "Response Model is off."}


@app.put("/tasks", response_model=TaskOut)
def create_task(task: TaskIn):
    return {"task_id": 1, "name": task.name, "description": task.description}


if __name__ == "__main__":
    app.run(host="localhost", port=5000)



"""Validation options with constraints."""

from flask_jeroboam import Jeroboam, Query

app = Jeroboam("Jeroboam Inbound Features App")
app.init_app()


@app.get("/argument_with_validation")
def get_validation_option(page: int = Query(ge=1)):
    return f"Received page argument is : {page}"


if __name__ == "__main__":
    app.run(host="localhost", port=5000)



"""Inbound Models for Testing."""

from pydantic import Field, field_validator

from flask_jeroboam import InboundModel


class SimpleModelIn(InboundModel):
    """Simple InboundModel for Testing Parsing Request."""

    page: int
    type: str


class QueryStringWithList(InboundModel):
    """A List of ids."""

    id: list[int] = Field(alias="id[]")


class OptionalModelIn(InboundModel):
    """a BaseModel with Optional Fields."""

    page: int | None = None
    per_page: int | None = None


class ModelWithListIn(InboundModel):
    """InboundModel with lists."""

    page: int
    per_page: int
    ids: list[int] = Field(alias="id[]")
    order: list[dict] = Field(alias="order[]")

    @field_validator("order")
    @classmethod
    def order_validator(cls, value):
        """Validate order."""
        if len(value) == 0:
            raise ValueError("Order must have at least 1 value")
        return value



"""View Parameters.

Subclasses of pydantic.fields.FieldInfo that are used to define
localised fields with some extra information.
"""

from enum import Enum
from typing import Any

from pydantic.fields import FieldInfo
from pydantic_core import PydanticUndefined


class ArgumentLocation(Enum):
    """Enum for the possible source location of a view_function parameter."""

    query = "query"
    header = "header"
    path = "path"
    cookie = "cookie"
    body = "body"
    form = "form"
    file = "file"
    unknown = "unknown"


class ViewArgument(FieldInfo):  # type: ignore[misc]
    """Base class for all View parameters.

    Inherits from pydantic v2 FieldInfo so that constraints (gt, lt,
    min_length, …) are stored natively and can be applied via
    TypeAdapter(Annotated[annotation, view_arg_instance]).
    """

    location: ArgumentLocation

    def __init__(
        self,
        default: Any = PydanticUndefined,
        **kwargs: Any,
    ):
        # Normalise Ellipsis (used as "required" marker in pydantic v1 style)
        if default is Ellipsis:
            default = PydanticUndefined
        self.embed = kwargs.pop("embed", False)
        self.include_in_schema = kwargs.pop("include_in_schema", True)
        self.example = kwargs.pop("example", PydanticUndefined)
        self.examples = kwargs.pop("examples", None)
        super().__init__(default=default, **kwargs)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.default})"

    @property
    def in_body(self):
        """Is the parameter located in the body?"""
        return self.location in {
            ArgumentLocation.body,
            ArgumentLocation.form,
            ArgumentLocation.file,
        }


class ParameterArgument(ViewArgument):  # type: ignore[misc]
    """A Parameter that is not located in the body."""

    def __init__(
        self,
        default: Any = PydanticUndefined,
        **kwargs: Any,
    ):
        self.deprecated = kwargs.pop("deprecated", None)
        super().__init__(
            default,
            **kwargs,
        )


class QueryArgument(ParameterArgument):  # type: ignore[misc]
    """A Parameter found in the Query String."""

    location = ArgumentLocation.query


class PathArgument(ParameterArgument):  # type: ignore[misc]
    """A Parameter found in Path."""

    location = ArgumentLocation.path

    def __init__(
        self,
        *args: Any,
        **kwargs: Any,
    ):
        # Path params are always required — ignore any supplied default
        super().__init__(
            PydanticUndefined,
            **kwargs,
        )


class HeaderArgument(ParameterArgument):  # type: ignore[misc]
    """A Header parameter."""

    location = ArgumentLocation.header

    def __init__(
        self,
        default: Any = PydanticUndefined,
        **kwargs: Any,
    ):
        self.convert_underscores = kwargs.pop("convert_underscores", True)
        super().__init__(
            default,
            **kwargs,
        )


class CookieArgument(ParameterArgument):  # type: ignore[misc]
    """A Parameter located in Cookies."""

    location = ArgumentLocation.cookie


class BodyArgument(ViewArgument):  # type: ignore[misc]
    """A Parameter located in Body.

    Body Parameters can be embedded. which means that they must
    be accessed by their name at the root of the body.
    They also have a Media/Type that varies between body, form and file.
    """

    location = ArgumentLocation.body

    def __init__(
        self,
        default: Any = PydanticUndefined,
        **kwargs: Any,
    ):
        self.media_type = kwargs.pop("media_type", "application/json")
        super().__init__(
            default,
            **kwargs,
        )


class FormArgument(BodyArgument):  # type: ignore[misc]
    """A Parameter located in Body."""

    location = ArgumentLocation.form

    def __init__(
        self,
        default: Any = PydanticUndefined,
        **kwargs: Any,
    ):
        embed = kwargs.pop("embed", True)
        kwargs.setdefault("media_type", "application/x-www-form-urlencoded")
        super().__init__(
            default,
            embed=embed,
            **kwargs,
        )


class FileArgument(FormArgument):  # type: ignore[misc]
    """A Parameter located in Body."""

    location = ArgumentLocation.file

    def __init__(
        self,
        default: Any = PydanticUndefined,
        **kwargs: Any,
    ):
        embed = kwargs.pop("embed", False)
        kwargs.setdefault("media_type", "multipart/form-data")
        super().__init__(
            default,
            embed=embed,
            **kwargs,
        )


def get_argument_class(location: ArgumentLocation) -> type[ViewArgument]:
    """Get the Parameter class for a given location."""
    return {
        ArgumentLocation.query: QueryArgument,
        ArgumentLocation.header: HeaderArgument,
        ArgumentLocation.path: PathArgument,
        ArgumentLocation.cookie: CookieArgument,
        ArgumentLocation.body: BodyArgument,
        ArgumentLocation.form: FormArgument,
        ArgumentLocation.file: FileArgument,
    }[location]



"""Flask-Jeroboam Blueprint.

Subclasses the Flask Blueprint.
Can be used as a drop in replacement of any Flask's Blueprint.
"""

from typing import Any

from flask.blueprints import Blueprint as FlaskBlueprint

from flask_jeroboam.scaffold import JeroboamScaffoldOverRide


class Blueprint(JeroboamScaffoldOverRide, FlaskBlueprint):  # type:ignore
    """Regular Blueprint with extra behavior on route definition."""

    def __init__(
        self,
        *args: Any,
        tags: list[str] | None = None,
        include_in_openapi: bool = True,
        **kwargs: Any,
    ) -> None:
        """Init."""
        self.include_in_openapi = include_in_openapi
        self.tags = tags or []
        super().__init__(*args, **kwargs)



"""List response - returning multiple items."""

from pydantic import BaseModel

from flask_jeroboam import Jeroboam

app = Jeroboam("Jeroboam Outbound Features App")
app.init_app()


class Task(BaseModel):
    id: int
    title: str
    done: bool = False


@app.get("/tasks", response_model=list[Task])
def list_tasks():
    return [
        {"id": 1, "title": "Buy groceries", "done": False},
        {"id": 2, "title": "Write docs", "done": True},
        {"id": 3, "title": "Review PR", "done": False},
    ]


if __name__ == "__main__":
    app.run(host="localhost", port=5000)



from pydantic import Field
from pydantic_settings import BaseSettings

from flask_jeroboam.openapi.models.openapi import Server


class JeroboamConfig(BaseSettings):
    """Jeroboam Config."""

    JEROBOAM_REGISTER_OPENAPI: bool | None = Field(default=True)

    JEROBOAM_TITLE: str | None = Field(None)
    JEROBOAM_VERSION: str | None = Field("0.1.0")
    JEROBOAM_DESCRIPTION: str | None = Field(None)
    JEROBOAM_TERMS_OF_SERVICE: str | None = Field(None)
    JEROBOAM_CONTACT: str | None = Field(None)
    JEROBOAM_LICENCE_INFO: str | None = Field(None)
    JEROBOAM_OPENAPI_VERSION: str | None = Field("3.0.2")
    JEROBOAM_SERVERS: list[Server] | None = Field([])
    JEROBOAM_OPENAPI_URL: str | None = Field(default="/docs")

    JEROBOAM_REGISTER_ERROR_HANDLERS: bool | None = Field(default=True)

    @classmethod
    def load(cls) -> "JeroboamConfig":
        """Load config."""
        return cls.model_validate({})



"""Augmented Rule for Flask-Jeroboam."""

import re
from typing import Any

from werkzeug.routing import Rule as FlaskRule

pattern = re.compile(r"<(\w*):")


class JeroboamRule(FlaskRule):
    """Subclass of werkzeug.routing.Rule."""

    def __init__(self, rule: str, endpoint: str | None = None, **options: Any) -> None:
        """Initialize a JeroRule."""
        self.include_in_openapi = endpoint != "static" and options.pop(
            "include_in_openapi", True
        )
        self.tags = options.pop("tags", [])

        # TODO: Maybe a smarter way to extract options...
        self.description = options.pop("description", None)
        self._summary = options.pop("summary", None)
        self.operation_id = options.pop("operation_id", None)
        self.deprecated = options.pop("deprecated", False)
        self.openapi_extra: dict = options.pop("openapi_extra", {})
        main_method = options.pop("main_method", "get")
        super().__init__(rule, endpoint=endpoint, **options)
        self.unique_id = options.pop(
            "unique_id", _generate_unique_id(self, main_method)
        )

    @property
    def openapi_path(self) -> str:
        """Return a formatted path."""
        rule = pattern.sub("<", self.rule)
        return rule.replace("<", "{").replace(">", "}")

    @property
    def summary(self) -> str:
        """Generate summary for the Rule.

        # TODO: Choper la description/summary dans la docstring de la view_function
        """
        return self._summary or self.endpoint.replace("_", " ").title().split(".")[-1]


def _generate_unique_id(rule: "JeroboamRule", main_method: str) -> str:
    def _split(my_string: str):
        return re.sub(r"[\W]+", "_", my_string).split("_")

    view_function_name = rule.endpoint.split(".")[-1]
    blueprints = ".".join(rule.endpoint.split(".")[:-1])
    http_verb = f"{main_method.lower()}"
    words = (
        _split(blueprints)
        + [None if http_verb in view_function_name else http_verb]
        + _split(view_function_name)
    )
    return "_".join([word for word in words if word])



"""The Route Class."""

from typing import Any, TypeVar

from typing_extensions import ParamSpec

from flask_jeroboam._inboundhandler import InboundHandler
from flask_jeroboam._outboundhandler import OutboundHandler
from flask_jeroboam.responses import JSONResponse
from flask_jeroboam.typing import JeroboamRouteCallable
from flask_jeroboam.view_arguments.solved import SolvedArgument

P = ParamSpec("P")
R = TypeVar("R")


class JeroboamView:
    """Adds flask-jeroboam features to a regular flask view function.

    The InboundHandler and OutboundHandler are configured here.
    The as_view property returns an augmented view function ready to be used by Flask,
    adding inbound and outbound data handling behavior if needed.
    The resulting view function is a regular flask view function, and any overhead
    related to figuring out what needs to be done is spent at registration time.
    """

    def __init__(
        self,
        rule: str,
        original_view_func: JeroboamRouteCallable,
        options: dict[str, Any],
        response_class: type | None = JSONResponse,
    ):
        """Initialize the JeroboamView.

        TODO: Is there a better way to set the default value of response_class?
        """
        assert response_class is not None  # noqa: S101
        self.endpoint = options.pop("endpoint", None)
        self.main_http_verb = self._solve_main_http_verb(options, original_view_func)
        configured_status_code: int | None = options.pop("status_code", None)
        self.inbound_handler = InboundHandler(
            original_view_func, self.main_http_verb, rule
        )
        self.outbound_handler = OutboundHandler(
            original_view_func,
            configured_status_code,
            self.main_http_verb,
            options,
            response_class,
        )
        self.original_view_func = original_view_func
        self.include_in_openapi = options.pop("include_in_openapi", True)
        self.has_request_body = self.inbound_handler.has_request_body
        self.rule = rule

    @property
    def as_view(self) -> JeroboamRouteCallable:
        """Decorate the orinal view function with handlers..

        TODO: Deal with name, docs, before decorating
        """
        view_func = self.original_view_func
        name = view_func.__name__
        doc = view_func.__doc__

        if self.inbound_handler.is_valid:
            view_func = self.inbound_handler.add_inbound_handling_to(view_func)
        view_func = self.outbound_handler.add_outbound_handling_to(view_func)

        view_func.__name__ = name
        view_func.__doc__ = doc
        # TODO voir comment passer l'objet lui-même.
        view_func.__jeroboam_view__ = self  # type: ignore
        return view_func

    @property
    def main_method(self) -> str:
        """Return the main HTTP verb of the Endpoint."""
        return self.main_http_verb.lower()

    @property
    def parameters(self) -> list[SolvedArgument]:
        """Return the main HTTP verb of the Endpoint."""
        return self.inbound_handler.parameters

    def _solve_main_http_verb(
        self, options: dict[str, Any], original_view_func: JeroboamRouteCallable
    ) -> str:
        """Return the main HTTP verb of the Endpoint.

        Here we consider that the main verb is either OPTIONS if it's the only
        one present,
        or either GET, PUT, POST, DELETE, PATCH, HEAD when present with OPTIONS.
        The presence of more than one verb from the latter list will raise
        a UserWarning.
        Only the fist one will be considered as the main verb.
        Default to GET if no verb is present.
        """
        methods = options.get("methods", ["GET"])
        if len(methods) == 1:
            return methods[0]
        if "OPTIONS" in methods:
            methods.remove("OPTIONS")
        if "HEAD" in methods:
            methods.remove("HEAD")
        if len(methods) >= 2:
            import warnings

            main_verbs = ", ".join(methods)
            warnings.warn(
                f"More than two primary HTTP verbs ({main_verbs}) are present"
                "on the endpoint "
                f"'{original_view_func.__module__}:{original_view_func.__name__}'. "
                f"The main verb is not clear and only the first one ({methods[0]}) "
                "will be picked. If you thing this is a mistake, please open an issue.",
                UserWarning,
                stacklevel=2,
            )
        return methods[0]



from flask_jeroboam.jeroboam import Jeroboam


def test_ui(client):
    """Test info endpoint."""
    response = client.get("/docs")
    assert response.status_code == 200
    assert response.data.startswith(b"<!DOCTYPE html>")


def test_opt_out_registering():
    """GIVEN a decorated view function
    WHEN it raises a ServerError
    THEN I get a 500 response with ServerError Message
    """
    app = Jeroboam(__name__)
    app.config["JEROBOAM_REGISTER_OPENAPI"] = False
    app.init_app()

    client = app.test_client()
    response = client.get("/docs")
    assert response.status_code == 404



"""Type patterns with Pydantic models."""

from pydantic import BaseModel

from flask_jeroboam import Jeroboam, Query

app = Jeroboam("Jeroboam Inbound Features App")
app.init_app()


class Item(BaseModel):
    name: str
    count: int


@app.get("/defining_type_with_type_hints")
def get_type_hints(page: int, search: list[str], item: Item, price=Query(float)):
    return (
        f"Received arguments are :\n"
        f"page : {page}\n"
        f"search : {search}\n"
        f"price : {price}\n"
        f"item : {item}"
    )


if __name__ == "__main__":
    app.run(host="localhost", port=5000)



"""Adding view arguments."""

from flask_jeroboam import Jeroboam

app = Jeroboam("Jeroboam Getting Started App")
app.init_app()


@app.get("/query_string_arguments")
def get_query_string_arguments(page: int = 1, per_page: int = 10):
    return {"page": page, "per_page": per_page}


if __name__ == "__main__":
    app.run(host="localhost", port=5000)



"""A Test Blueprint for testing Body Params.

The corresponding test can be found in tests/test_inbound/test_body
"""

from flask_jeroboam import Body
from flask_jeroboam.blueprint import Blueprint
from tests.app_test.models.inbound import SimpleModelIn

router = Blueprint("body_params_router", __name__, tags=["Body"])


@router.post("/body/int")
def post_body_as_int(payload: int = Body()):
    """Body Param as plain int."""
    return {"payload": payload}


@router.post("/body/str")
def post_body_as_str(payload: str = Body()):
    """Body Param as plain str."""
    return {"payload": payload}


@router.post("/body/base_model")
def post_base_model_in_form(payload: SimpleModelIn = Body()):
    """POST Form Parameter as pydantic BaseModel."""
    return payload.model_dump_json()


@router.post("/body/multi_primitive")
def post_multi_primitive_body_arguments(
    age: int = Body(media_type="application/json"),
    name: str = Body(media_type="application/ujson"),
):
    return {"age": age, "name": name}



"""Outbound Models for Testing."""

from dataclasses import dataclass

from flask_jeroboam import OutboundModel


class SimpleModelOut(OutboundModel):
    """Base OutBoundModel for Testing."""

    total_count: int
    items: list[str]


@dataclass
class MyDataClass:
    """A Simple DataClass for Testing."""

    total_count: int
    items: list[str]


class UserOut(OutboundModel):
    """Only the username must be returned."""

    username: str


class UserIn(UserOut):
    """Inbound contains the password."""

    password: str


class ModelWithListOut(OutboundModel):
    """OutboundModel with lists."""

    page: int
    per_page: int
    ids: list[int]
    order: list[dict]



"""DataStructures.

Credits: this module is essentially a for of FastAPI's datastructures.py module.
"""

from typing import Any

from pydantic import GetCoreSchemaHandler
from pydantic_core import CoreSchema, core_schema
from werkzeug.datastructures import FileStorage


class UploadFile(FileStorage):
    """A wrapper around werkzeug.datastructures.FileStorage.

    Credits: Adaptation of FastAPI's UploadFile.
    """

    @classmethod
    def validate(cls: type["FileStorage"], v: Any) -> Any:
        """Validate the value."""
        if not isinstance(v, FileStorage):
            raise ValueError(f"Expected FileStorage, received: {type(v)}")
        return v

    @classmethod
    def __get_pydantic_core_schema__(
        cls, source_type: Any, handler: GetCoreSchemaHandler
    ) -> CoreSchema:
        """Pydantic v2 schema."""
        return core_schema.no_info_plain_validator_function(
            cls.validate,
            serialization=core_schema.plain_serializer_function_ser_schema(str),
        )

    @classmethod
    def __get_pydantic_json_schema__(cls, schema: Any, handler: Any) -> dict[str, Any]:
        return {"type": "string", "format": "binary"}



"""The Flask Object with augmented functionality around route registration.

Here we overide the route method of the Flask object to use our custom implementation.
This allow us to introduce new functionality to the route registration process.

TODO: A probably better way to override it is to override the url_rule_class
with a custom JeroboamRule Object
"""

from collections.abc import Callable
from typing import Any, Optional

from flask import Flask
from typing_extensions import TypeVar

from flask_jeroboam._config import JeroboamConfig
from flask_jeroboam.exceptions import register_error_handlers
from flask_jeroboam.openapi.blueprint import register_open_api_blueprint
from flask_jeroboam.openapi.builder import build_openapi
from flask_jeroboam.openapi.models.openapi import OpenAPI
from flask_jeroboam.responses import JSONResponse
from flask_jeroboam.rule import JeroboamRule
from flask_jeroboam.scaffold import JeroboamScaffoldOverRide

R = TypeVar("R", bound=Any)


class Jeroboam(JeroboamScaffoldOverRide, Flask):  # type:ignore
    """A Flask Object with extra functionalities.

    The route method is overriden by a custom flask_jeroboam
    route decorator.
    """

    response_class = JSONResponse

    url_rule_class = JeroboamRule

    query_string_key_transformer: Callable | None = None

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        """Init."""
        super().__init__(*args, **kwargs)
        self.config.update(JeroboamConfig.load().model_dump())
        self._openapi: OpenAPI | None = None

    def init_app(self, app: Optional["Jeroboam"] = None) -> None:
        """Setup is performed after app has received all its configuration."""
        if self.config["JEROBOAM_REGISTER_ERROR_HANDLERS"]:
            register_error_handlers(self)  # type: ignore
        if self.config["JEROBOAM_REGISTER_OPENAPI"]:
            register_open_api_blueprint(self)  # type: ignore

    @property
    def openapi(self) -> OpenAPI:
        """Get the OpenApi object."""
        if self._openapi is None:
            self._openapi = build_openapi(
                app=self,  # type: ignore
                rules=list(self.url_map.iter_rules()),  # type: ignore
                tags=[],
            )
        return self._openapi



import pytest

not_a_valid_int = {
    "detail": [
        {
            "loc": ["header", "Test-Header"],
            "msg": "Input should be a valid integer, unable to parse string as an integer",
            "type": "int_parsing",
        }
    ]
}


def _valid(x):
    return {"header": x}


@pytest.mark.parametrize(
    "url,header_value,expected_status,expected_response",
    [
        ("/headers/str", {"test-header": "foobar"}, 200, _valid("foobar")),
        ("/headers/int", {"test-header": "123"}, 200, _valid(123)),
        ("/headers/int", {"test-header": "not_a_valid_int"}, 400, not_a_valid_int),
    ],
)
def test_get_headers(client, url, header_value, expected_status, expected_response):
    """Test Cookie Parameter with GET method.


    TODO: Allow Configuration of the returned Status Code.
    """
    response = client.get(url, headers=header_value)
    assert response.status_code == expected_status
    assert response.json == expected_response



import pytest

"""Testing the inbound feature doc Page."""
from docs_src.features.inbound import app


@pytest.fixture()
def client():
    """A Test Client for the app."""
    return app.test_client()


def test_implicit_location_on_get(client):
    """Test the implicit location GET example."""
    response = client.get("/implicit_location_is_query_string?page=42")
    assert response.data == b"Received Page Argument is : 42"
    assert response.status_code == 200


def test_implicit_location_on_post(client):
    """Test the implicit location POTS example."""
    response = client.post("/implicit_location_is_body", json={"page": 42})
    assert response.data == b"Received Page Argument is : 42"

    assert response.status_code == 201


def test_implicit_path_parameter(client):
    """Test the implicit path parameter example."""
    response = client.get("/item/42/implicit")
    assert response.data == b"Received id Argument is : 42"
    assert response.status_code == 200


def test_implicit_post_path_parameter(client):
    """Test the implicit path parameter example."""
    response = client.post("/item/42/implicit")
    assert response.data == b"Received id Argument is : 42"
    assert response.status_code == 201


def test_explicit_location_on_get(client):
    """Test the explicit location GET example."""
    response = client.get("/explicit_location_is_query_string?page=42")
    assert response.data == b"Received Page Argument is : 42"
    assert response.status_code == 200


def test_explicit_location_on_post(client):
    """Test the explicit location GET example."""
    response = client.post("/explicit_location_is_body", json={"page": 42})
    assert response.data == b"Received Page Argument is : 42"
    assert response.status_code == 201


def test_explicit_cookie_location_with_other_explicit(client):
    """Test the explicit location GET example."""
    client.set_cookie("username", "john")
    response = client.get(
        "/explicit_location_is_query_string_and_cookie?page=42",
    )
    assert response.data == b"Received Page Argument is : 42. Username is : john"
    assert response.status_code == 200


def test_explicit_cookie_location_with_other_implicit(client):
    """Test the explicit location GET example."""
    client.set_cookie("username", "john")
    response = client.get(
        "/implicit_and_explicit?page=42",
    )
    assert response.data == b"Received Page Argument is : 42. Username is : john"
    assert response.status_code == 200


def test_argument_is_required(client):
    """Test the explicit location GET example."""
    client.set_cookie("username", "john")
    response = client.get(
        "implicit_location_is_query_string",
    )
    assert response.json == {
        "detail": [
            {
                "loc": ["query", "page"],
                "msg": "Field required",
                "type": "missing",
            }
        ]
    }
    assert response.status_code == 400


def test_implicit_default_values(client):
    """Test default values with implicit-location."""
    response = client.get(
        "implicit_location_with_default_value",
    )
    assert response.data == b"Received Page Argument is : 1"
    assert response.status_code == 200


def test_explicit_default_values(client):
    """Test default values with implicit-location."""
    response = client.get(
        "explicit_location_with_default_value",
    )
    assert response.data == b"Received Page Argument is : 1"
    assert response.status_code == 200


def test_implicit_default_values_get_injected(client):
    """Test default values with implicit-location."""
    response = client.get(
        "implicit_location_with_default_value?page=42",
    )
    assert response.data == b"Received Page Argument is : 42"
    assert response.status_code == 200


def test_explicit_default_values_get_injected(client):
    """Test default values with implicit-location."""
    response = client.get(
        "explicit_location_with_default_value?page=42",
    )
    assert response.data == b"Received Page Argument is : 42"
    assert response.status_code == 200


def test_various_type_definition(client):
    """Test default values with implicit-location."""
    response = client.get(
        "defining_type_with_type_hints?page=42&search=foo&search=bar&price=42.42&name=test&count=3",
    )
    assert response.data == (
        b"Received arguments are :\n"
        b"page : 42\n"
        b"search : ['foo', 'bar']\n"
        b"price : 42.42\n"
        b"item : name='test' count=3"
    )
    assert response.status_code == 200


def test_validation_option(client):
    """Test default values with implicit-location."""
    response = client.get("argument_with_validation?page=0")
    assert response.json == {
        "detail": [
            {
                "ctx": {"ge": 1.0},
                "loc": ["query", "page"],
                "msg": "Input should be greater than or equal to 1",
                "type": "greater_than_equal",
            }
        ]
    }
    assert response.status_code == 400



from flask.testing import FlaskClient
from pydantic import BaseModel, field_validator

from flask_jeroboam.jeroboam import Jeroboam


class ModelB(BaseModel):
    username: str


class ModelC(ModelB):
    password: str


class ModelA(BaseModel):
    name: str
    description: str | None = None
    model_b: ModelB

    @field_validator("name")
    @classmethod
    def lower_username(cls, name: str):
        """Validate that the name ends in A."""
        if not name.endswith("A"):
            raise ValueError("name must end in A")
        return name


def test_get_query_operations(one_shot_app: Jeroboam, one_shot_client: FlaskClient):
    """Testing Various GET operations with query parameters.

    GIVEN a GET endpoint configiured with query parameters
    WHEN a request is made to the endpoint
    THEN the request is parsed and validated accordingly
    """

    @one_shot_app.post("/sub_model", response_model=ModelA)
    def post_sub_model(model: ModelA):
        return model

    response = one_shot_client.post(
        "/sub_model", json={"name": "fooA", "model_b": {"username": "bar"}}
    )
    assert response.status_code == 201
    assert response.json == {
        "name": "fooA",
        "model_b": {"username": "bar"},
        "description": None,
    }



"""Code Example 00 for README.md."""

from flask_jeroboam import Jeroboam

app = Jeroboam(__name__)


@app.get("/ping")
def ping():
    return "pong"


if __name__ == "__main__":
    app.run(port=5000)



"""Solved Specialised Params.

Params are solved at registration time. This way we reduce indirections when
handling requests thus reducing overhead.
"""

import re
from copy import deepcopy
from typing import Annotated, Any

from flask import request
from pydantic import TypeAdapter, ValidationError
from pydantic_core import ErrorDetails, PydanticUndefined
from werkzeug.datastructures import FileStorage, Headers, MultiDict

from flask_jeroboam._utils import _unwrap_optional
from flask_jeroboam.view_arguments._utils import (
    _extract_scalar,
    _extract_sequence,
    _extract_subfields,
)
from flask_jeroboam.view_arguments.arguments import ArgumentLocation, ViewArgument


class SolvedArgument:
    """Generic Solved Parameter.

    Created at route-registration time; reused on every request.
    Validates inbound data using a pydantic v2 TypeAdapter.
    """

    def __init__(
        self,
        *,
        name: str,
        annotation: type,
        required: bool = False,
        default: Any = PydanticUndefined,
        location: ArgumentLocation = ArgumentLocation.unknown,
        alias: str | None = None,
        embed: bool = False,
        include_in_schema: bool = True,
        field_info: ViewArgument | None = None,
    ):
        self.name = name
        self.annotation = annotation
        self.required = required
        self.default = default
        self.location = location
        self.alias = alias or name
        self.embed = embed
        self.include_in_schema = include_in_schema
        self.field_info = field_info  # the original ViewArgument

        # Build TypeAdapter once at registration time.
        # Wrap in Annotated so FieldInfo constraints (gt, lt, …) are applied.
        if field_info is not None:
            adapter_type: Any = Annotated[annotation, field_info]
        else:
            adapter_type = annotation
        self._type_adapter: TypeAdapter = TypeAdapter(adapter_type)

    @classmethod
    def specialize(
        cls,
        *,
        name: str,
        annotation: type,
        required: bool = False,
        view_param: ViewArgument | None = None,
        **_kwargs,
    ) -> "SolvedArgument":
        """Dispatch to the location-specific subclass."""
        location = getattr(view_param, "location", ArgumentLocation.unknown)
        target_class = {
            ArgumentLocation.query: SolvedQueryArgument,
            ArgumentLocation.header: SolvedHeaderArgument,
            ArgumentLocation.path: SolvedPathArgument,
            ArgumentLocation.cookie: SolvedCookieArgument,
            ArgumentLocation.body: SolvedBodyArgument,
            ArgumentLocation.file: SolvedFileArgument,
            ArgumentLocation.form: SolvedFormArgument,
        }.get(location, cls)

        default = getattr(view_param, "default", PydanticUndefined)
        if default is Ellipsis:
            default = PydanticUndefined

        return target_class(
            name=name,
            annotation=annotation,
            required=required,
            default=default,
            location=location,
            alias=getattr(view_param, "alias", None),
            embed=getattr(view_param, "embed", False),
            include_in_schema=getattr(view_param, "include_in_schema", True),
            field_info=view_param,
        )

    def validate_request(self) -> tuple[dict, list[dict]]:
        """Extract and validate the parameter from the current request."""
        values: dict = {}
        errors: list[dict] = []

        inbound_values = self._get_values()

        if inbound_values is None and self.required:
            errors.append(
                {
                    "loc": [self.location.value, self.alias],
                    "msg": "Field required",
                    "type": "missing",
                }
            )
            return values, errors

        if inbound_values is None:
            if self.default is not PydanticUndefined:  # pragma: no branch
                values[self.name] = deepcopy(self.default)
            return values, errors

        try:
            values[self.name] = self._type_adapter.validate_python(inbound_values)
        except ValidationError as exc:
            errors.extend(
                self._format_error(err) for err in exc.errors(include_url=False)
            )

        return values, errors

    def _format_error(self, err: ErrorDetails) -> dict:
        """Format a pydantic ValidationError entry into the Jeroboam error shape."""
        loc = [self.location.value, self.alias] + [
            str(segment) for segment in err.get("loc", ())
        ]
        error_dict: dict = {"loc": loc, "msg": err["msg"], "type": err["type"]}
        if "ctx" in err:
            error_dict["ctx"] = {
                k: str(v) if isinstance(v, Exception) else v
                for k, v in err["ctx"].items()
            }
        return error_dict

    def _get_values(
        self,
    ) -> FileStorage | MultiDict | dict | str | None | list[Any]:
        """The Value extraction method is specialised by location."""
        raise NotImplementedError


class SolvedPathArgument(SolvedArgument):
    """Solved Path parameter."""

    def _get_values(self) -> dict | str | None | list[Any]:
        source: dict = request.view_args or {}
        return _extract_scalar(source=source, alias=self.alias, name=self.name)


class SolvedHeaderArgument(SolvedArgument):
    """Solved Header parameter."""

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        # Compute the HTTP header name from the Python field name.
        # e.g. content_type → Content-Type
        self.alias = re.sub(
            r"_(\w)", lambda x: f"-{x.group(1).upper()}", self.name.capitalize()
        )

    def _get_values(self) -> dict | str | None | list[Any]:
        source: Headers | dict = request.headers or {}
        return _extract_scalar(source=source, alias=self.alias, name=self.name)


class SolvedCookieArgument(SolvedArgument):
    """Solved Cookie parameter."""

    def _get_values(self) -> dict | str | None | list[Any]:
        source: dict = request.cookies or {}
        return _extract_scalar(source=source, alias=self.alias, name=self.name)


class SolvedQueryArgument(SolvedArgument):
    """Solved Query parameter."""

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        from typing import get_origin

        # Unwrap Optional[X] once at registration time; reuse on every request.
        self._inner = _unwrap_optional(self.annotation)
        if hasattr(self._inner, "model_fields"):
            self._fields = self._inner.model_fields
            self.extractor = _extract_subfields
        elif get_origin(self._inner) in (list, tuple, set, frozenset):
            self._fields = {}
            self.extractor = _extract_sequence
        else:
            self._fields = {}
            self.extractor = _extract_scalar

    def _get_values(self) -> dict | str | None | list[Any]:
        source: MultiDict = request.args
        return self.extractor(
            source=source,
            alias=self.alias,
            name=self.name,
            fields=self._fields,
        )


class SolvedBodyArgument(SolvedArgument):
    """Solved Body parameter."""

    def _get_values(self) -> dict | str | None | list[Any]:
        source: dict | list = request.json or {}
        if isinstance(source, list):
            return source
        return source.get(self.alias or self.name) if self.embed else source


class SolvedFileArgument(SolvedArgument):
    """Solved File Parameter."""

    def _get_values(
        self,
    ) -> FileStorage | MultiDict | dict | str | None | list[Any]:
        source: MultiDict = request.files or MultiDict()
        return source.get(self.alias or self.name) if self.embed else source


class SolvedFormArgument(SolvedArgument):
    """Solved Form parameter."""

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self._inner = _unwrap_optional(self.annotation)
        if not self.embed and hasattr(self._inner, "model_fields"):
            self._fields = self._inner.model_fields
            self._extractor = _extract_subfields
        else:
            self._fields = {}
            self._extractor = None

    def _get_values(self) -> dict | str | None | list[Any]:
        source: MultiDict = request.form or MultiDict()
        if self.embed:
            return source.get(self.alias or self.name)
        if self._extractor is not None:
            return self._extractor(
                source=source,
                alias=self.alias,
                name=self.name,
                fields=self._fields,
            )
        return source  # pragma: no cover



"""Testing Examples from the Getting Started Page"""

from docs_src.getting_started_00 import app


def test_getting_started_00(app=app):
    """Test the first example from the Getting Started Page."""
    client = app.test_client()
    response = client.get("/health")
    assert response.status_code == 200
    assert response.json == {"status": "Ok"}


def test_getting_started_01(app=app):
    """Test the second example from the Getting Started Page."""
    client = app.test_client()
    response = client.get("/query_string_arguments")
    assert response.status_code == 200
    assert response.json == {"page": 1, "per_page": 10}


def test_getting_started_02(app=app):
    """Test the third example from the Getting Started Page."""
    client = app.test_client()
    response = client.get("/query_string_arguments?page=2&per_page=50")
    assert response.status_code == 200
    assert response.json == {"page": 2, "per_page": 50}


def test_getting_started_03(app=app):
    """Test the fourth example from the Getting Started Page."""
    client = app.test_client()
    response = client.get("/query_string_arguments?page=not_a_int&per_page=50")
    assert response.status_code == 400
    assert response.json == {
        "detail": [
            {
                "loc": ["query", "page"],
                "msg": "Input should be a valid integer, unable to parse string as an integer",
                "type": "int_parsing",
            }
        ]
    }


def test_getting_started_04(app=app):
    """Test the fifth example from the Getting Started Page."""
    client = app.test_client()
    response = client.get("/item")
    assert response.status_code == 200
    assert response.json == {"name": "Bottle", "price": 5.0, "count": 1}


def test_getting_started_05(app=app):
    """Test the sixth example from the Getting Started Page."""
    client = app.test_client()
    response = client.get("/docs")
    assert response.status_code == 200
    assert response.data.startswith(b"<!DOCTYPE html>")



from flask.testing import FlaskClient


def test_valid_payload_in_data_is_injected(
    client: FlaskClient,
):
    """GIVEN a POST endpoint with a BaseModel FormParam
    WHEN hit with valid form data
    THEN the parsed input is injected into the view function.
    """
    response = client.post("/form/base_model", data={"page": 1, "type": "item"})

    assert response.json == {"page": 1, "type": "item"}
    assert response.status_code == 201


def test_single_primitive_form_param_is_injected(
    client: FlaskClient,
):
    """GIVEN a POST endpoint with a single primitive Form param (embed=True)
    WHEN hit with valid form data
    THEN the parsed value is injected into the view function.
    """
    response = client.post("/form/single_primitive", data={"name": "Alice"})

    assert response.json == {"name": "Alice"}
    assert response.status_code == 201


def test_single_primitive_form_param_uses_default(
    client: FlaskClient,
):
    """GIVEN a POST endpoint with a single primitive Form param with default
    WHEN hit without providing that field
    THEN the default value is used.
    """
    response = client.post("/form/single_primitive", data={})

    assert response.json == {"name": "John"}
    assert response.status_code == 201


def test_multiple_primitive_form_params_are_injected(
    client: FlaskClient,
):
    """GIVEN a POST endpoint with multiple primitive Form params
    WHEN hit with valid form data
    THEN each value is injected as an individual kwarg.
    """
    response = client.post(
        "/form/primitives_in_form", data={"name": "Alice", "age": 25}
    )

    assert response.status_code == 201
    assert response.json == {"name": "Alice", "age": 25}


def test_multiple_primitive_form_params_use_defaults(
    client: FlaskClient,
):
    """GIVEN a POST endpoint with multiple primitive Form params with defaults
    WHEN hit without providing those fields
    THEN the default values are used.
    """
    response = client.post("/form/primitives_in_form", data={})

    assert response.status_code == 201
    assert response.json == {"name": "John", "age": 34}



"""Response model - Task example."""

from pydantic import BaseModel

from flask_jeroboam import Jeroboam

app = Jeroboam("Jeroboam Outbound Features App")
app.init_app()


class Task(BaseModel):
    id: int
    title: str
    description: str | None = None
    done: bool = False


@app.get("/tasks/<int:task_id>", response_model=Task)
def get_task(task_id: int):
    return {
        "id": task_id,
        "title": "Buy groceries",
        "description": "Milk, cheese, bread",
        "done": False,
    }


if __name__ == "__main__":
    app.run(host="localhost", port=5000)
